from ripple_down_rules.datastructures.case import Case
from typing_extensions import Dict, List, Set, Union
from decorator_model import Robot
from relational_model import PhysicalObject


def conditions_167615852950279355863004646114673699744(case) -> bool:
    def conditions_for_robot_get_contained_objects(self_: Robot, output_: PhysicalObject) -> bool:
        """Get conditions on whether it's possible to conclude a value for Robot_get_contained_objects.output_  of type PhysicalObject."""
        return len(self_.parts) > 0
    return conditions_for_robot_get_contained_objects(**case)


def conclusion_167615852950279355863004646114673699744(case) -> List[PhysicalObject]:
    def robot_get_contained_objects(self_: Robot, output_: PhysicalObject) -> List[PhysicalObject]:
        """Get possible value(s) for Robot_get_contained_objects.output_  of type PhysicalObject."""
        contained_objects = []
        for part in self_.parts:
            contained_objects.extend(part.contained_objects)
        return contained_objects
    return robot_get_contained_objects(**case)





# This file contains the corner cases for the rules.
from ripple_down_rules.datastructures.dataclasses import CaseFactoryMetaData


corner_case_167615852950279355863004646114673699744 = CaseFactoryMetaData(factory_method=None, factory_idx=None, case_conf=None, scenario=None)





import pytest

from ripple_down_rules import *
from .datasets import Drawer, Handle


@pytest.fixture
def drawer_cabinet_dependency_graph():
    TrackedObjectMixin.make_class_dependency_graph(composition=True)


def test_fit_depends_on_predicate(drawer_cabinet_dependency_graph) -> None:
    dependsOn.rdr_decorator.fit = True
    dependsOn.rdr_decorator.update_existing_rules = False
    assert dependsOn(Drawer, Handle)



from __future__ import annotations

import enum
import importlib
import inspect
import logging
import sys
import typing
from dataclasses import dataclass, Field
from datetime import datetime
from functools import lru_cache
from types import NoneType

from typing_extensions import Type, get_origin, Optional, get_type_hints, Tuple

from ripple_down_rules.utils import make_tuple


class ParseError(TypeError):
    """
    Error that will be raised when the parser encounters something that can/should not be parsed.

    For instance, Union types
    """
    pass


@dataclass(unsafe_hash=True)
class FieldInfo:
    """
    A class that wraps a field of dataclass and provides some utility functions.
    """

    clazz: Type
    """
    The class that the field is in.
    """

    name: str
    """
    The name of the field.
    """

    type: Tuple[Type, ...]
    """
    The type of the field or inner type of the container if it is a container.
    """

    optional: bool
    """
    True if the field is optional, False otherwise.
    """

    container: Optional[Type]
    """
    The type of the container if it is one (list, set, tuple, etc.). If there is no container this is None
    """

    is_type_field: bool = False
    """
    The field value is a type.
    """

    field: Field = None
    """
    The field instance.
    """

    def __init__(self, clazz: Type, f: Field):
        self.field = f
        self.name = f.name
        self.clazz = clazz

        try:
            type_hints = get_type_hints(clazz)[self.name]
        except NameError as e:
            found_clazz = manually_search_for_class_name(e.name)
            module = importlib.import_module(found_clazz.__module__)
            locals()[e.name] = getattr(module, e.name)
            type_hints = get_type_hints(clazz, localns=locals())[self.name]
        type_args = typing.get_args(type_hints)

        # try to unpack the type if it is a nested type
        if len(type_args) > 0:
            self.optional = NoneType in type_args

            if self.optional:
                self.container = None
            else:
                self.container = get_origin(type_hints)

            if not self.optional and type_hints == Type[type_args]:
                self.is_type_field = True

            self.type = type_args
        else:
            self.optional = False
            self.container = None
            self.type = make_tuple(type_hints)

    @property
    def is_builtin_class(self) -> bool:
        return not self.container and all(t.__module__ == 'builtins' for t in self.type)

    @property
    def is_container_of_builtin(self) -> bool:
        return self.container and all(t.__module__ == 'builtins' for t in self.type)

    @property
    def is_type_type(self) -> bool:
        return self.is_type_field

    @property
    def is_enum(self):
        return len(self.type) == 1 and issubclass(self.type[0], enum.Enum)

    @property
    def is_datetime(self):
        return len(self.type) == 1 and self.type[0] == datetime


def is_container(clazz: Type) -> bool:
    """
    Check if a class is an iterable.

    :param clazz: The class to check
    :return: True if the class is an iterable, False otherwise
    """
    return get_origin(clazz) in [list, set, tuple]


def manually_search_for_class_name(target_class_name: str) -> Type:
    """
    Searches for a class with the specified name in the current module's `globals()` dictionary
    and all loaded modules present in `sys.modules`. This function attempts to find and resolve
    the first class that matches the given name. If multiple classes are found with the same
    name, a warning is logged, and the first one is returned. If no matching class is found,
    an exception is raised.

    :param target_class_name: Name of the class to search for.
    :return: The resolved class with the matching name.

    :raises ValueError: Raised when no class with the specified name can be found.
    """
    found_classes = []

    # Search 1: In the current module's globals()
    for name, obj in globals().items():
        if inspect.isclass(obj) and obj.__name__ == target_class_name:
            found_classes.append(obj)

    # Search 2: In all loaded modules (via sys.modules)
    for module_name, module in sys.modules.items():
        if module is None or not hasattr(module, '__dict__'):
            continue  # Skip built-in modules or modules without a __dict__

        for name, obj in module.__dict__.items():
            if inspect.isclass(obj) and obj.__name__ == target_class_name:
                # Avoid duplicates if a class is imported into multiple namespaces
                if (obj, f"from module '{module_name}'") not in found_classes:
                    found_classes.append(obj)

    # If you wanted to "resolve" the forward ref based on this
    if len(found_classes) == 0:
        raise ValueError(f"Could not find any class with name {target_class_name} in globals or sys.modules.")
    elif len(found_classes) == 1:
        resolved_class = found_classes[0]
    else:
        warn_multiple_classes(target_class_name, tuple(found_classes))
        resolved_class = found_classes[0]

    return resolved_class


@lru_cache(maxsize=None)
def warn_multiple_classes(target_class_name, found_classes):
    logging.warning(f"Found multiple classes with name {target_class_name}. Found classes: {found_classes} ")



from typing_extensions import List, Set, Union
from relational_model import PhysicalObject, Robot


def conditions_298609776593271728826836208156881692889(case) -> bool:
    def conditions_for_robot_contained_objects_of_type_physical_object(case: Robot) -> bool:
        """Get conditions on whether it's possible to conclude a value for Robot.contained_objects  of type PhysicalObject."""
        return len(case.parts) > 0
    return conditions_for_robot_contained_objects_of_type_physical_object(case)


def conclusion_298609776593271728826836208156881692889(case) -> List[PhysicalObject]:
    def robot_contained_objects_of_type_physical_object(case: Robot) -> List[PhysicalObject]:
        """Get possible value(s) for Robot.contained_objects  of type PhysicalObject."""
        contained_objects: List[PhysicalObject] = []
        for part in case.parts:
            contained_objects.extend(part.contained_objects)
        return contained_objects
    return robot_contained_objects_of_type_physical_object(case)





from relational_model import Part, Robot, PhysicalObject
from ripple_down_rules import GeneralRDR, CaseQuery
from ripple_down_rules.helpers import enable_gui

# Define a simple robot with parts and containment relationships
part_a = Part(name="A")
part_b = Part(name="B")
part_c = Part(name="C")
robot = Robot("pr2", parts=[part_a])
part_a.contained_objects = [part_b]
part_b.contained_objects = [part_c]

# Optional: Use the GUI.
enable_gui()

# Create a GeneralRDR instance and fit it to the case query
grdr = GeneralRDR(save_dir='./', model_name='part_containment_rdr')

case_query = CaseQuery(robot, "contained_objects", (PhysicalObject,), False)

grdr.fit_case(case_query)

# Classify the robot to check if it rcontains part_b
result = grdr.classify(robot)
assert list(result['contained_objects']) == [part_b]


import pytest

from examples.relational_model import Part, Robot, PhysicalObject
from ripple_down_rules import GeneralRDR, CaseQuery
from ripple_down_rules.helpers import enable_gui

enable_gui()

@pytest.fixture
def robot_rdr():
    return GeneralRDR(save_dir='./', model_name='robot_rdr')

def robot_factory():
    part_a = Part(name="A")
    part_b = Part(name="B")
    part_c = Part(name="C")
    robot = Robot("pr2", parts=[part_a])
    part_a.contained_objects = [part_b]
    part_b.contained_objects = [part_c]
    return robot

def test_robot_contained_objects(robot_rdr):
    robot = robot_factory()
    cq = CaseQuery(robot, "contained_objects", (PhysicalObject,), False,
                   scenario=test_robot_contained_objects, case_factory=robot_factory)
    robot_rdr.fit_case(cq, update_existing_rules=False)
    result = robot_rdr.classify(robot)
    assert list(result['contained_objects']) == [robot.parts[0].contained_objects[0]]  # part_b


"""
This file contains decorators for the RDR (Ripple Down Rules) framework. Where each type of RDR has a decorator
that can be used with any python function such that this function can benefit from the incremental knowledge acquisition
of the RDRs.
"""
import inspect
import os.path
from dataclasses import dataclass, field
from functools import wraps

from typing_extensions import Callable, Optional, Type, Tuple, Dict, Any, List

from .datastructures.dataclasses import CaseFactoryMetaData
from .experts import Expert, Human
from .rdr import GeneralRDR
from .utils import get_origin_type_of_function_output, get_arg_type_of_function_output, fill_in_missing_kwargs

try:
    from .user_interface.gui import RDRCaseViewer
except ImportError:
    RDRCaseViewer = None
from .utils import get_func_rdr_model_name, make_set, \
    make_list
from .helpers import create_case_from_method, create_case_query_from_method


@dataclass(unsafe_hash=True)
class RDRDecorator:
    models_dir: str
    """
    The directory to save the RDR models in.
    """
    output_type: Tuple[Type, ...]
    """
    The type(s) of the output produced by the RDR model (The type(s) of the queried attribute).
    """
    mutual_exclusive: bool
    """
    Whether the queried attribute is mutually exclusive (i.e. allows for only one possible value) or not.
    """
    fit: bool = field(default=True)
    """
    Whether to run in fitting mode and prompt the expert or just classify using existing rules.
    """
    expert: Optional[Expert] = field(default=None)
    """
    The expert instance, this is used by the rdr to prompt for answers.
    """
    update_existing_rules: bool = field(default=True)
    """
    When in fitting mode, whether to ask for answers for existing rules as well or not.
    """
    package_name: Optional[str] = field(default=None)
    """
    The name of the user python package where the RDR model will be saved, this is useful for generating relative
    imports in the generated rdr model files.
    """
    use_generated_classifier: bool = field(default=False)
    """
    Whether to use the generated classifier files of the rdr model instead of the RDR instance itself, this is useful
    when you want to debug inside the rules.
    """
    ask_now: Callable[Dict[str, Any], bool] = field(default=lambda _: True)
    """
    A user provided callable function that outputs a boolean indicating when to ask the expert for an answer. The input
    to the `ask_now` function is a dictionary with the original function arguments, while arguments like `self` and
    `cls` are passed as a special key `self_` or `cls_` respectively.
    """
    fitting_decorator: Optional[Callable] = field(default=lambda f: f)
    """
    A user provided decorator that wraps the `py:meth:ripple_down_rules.rdr.RippleDownRules.fit_case` method which is 
    used when in fitting mode, this is useful when you want special logic pre and post the fitting operation, you can
    for example freeze your system during fitting such that you have a stable state that you can query and use while 
    writing and testing your answer/rule.
    """
    generate_dot_file: bool = field(default=False)
    """
    Whether to generate a dynamic dot file representing the state of the rule tree each time the rdr is queried, showing
    which rules fired and which rules didn't get evaluated, ...etc.
    """
    model_name: Optional[str] = field(default=None)
    """
    The name of the rdr model, this gets auto generated from the function signature and the class/file it is contained
    in.
    """
    rdr: GeneralRDR = field(init=False, default=None)
    """
    The ripple down rules instance of the decorator class.
    """
    parsed_output_type: List[Type] = field(init=False, default_factory=list)
    """
    The output of a post processing done on the output types, for example converting typing module types 
    (i.e. type hints) to python types.
    """
    origin_type: Optional[Type] = field(init=False, default=None)
    """
    The origin of the type hint of the attribute, useful in the case of not mutually exclusive attributes to map the 
    result to the specified container type (e.g. a list instead of a set which is the default container type for rdr
    output).
    """
    output_name: str = field(init=False, default='output_')
    """
    This is used to refer to the output value of the decorated function, this is used as part of the case as input to 
    the rdr model, but is never used in the rule logic to prevent cycles from happening. The correct way to use the 
    output of an rdr is through refinement rules which happens automatically by the rdr prompting for refinements.
    """
    _not_none_output_found: bool = field(init=False, default=False)
    """
    This is a flag that indicates that a not None output for the rdr has been inferred, this is used to update the 
    generated dot file if it is set to `True`.
    """
    case_factory_metadata: CaseFactoryMetaData = field(init=False, default_factory=CaseFactoryMetaData)
    """
    Metadata that contains the case factory method, and the scenario that is being run during the case query.
    """

    def decorator(self, func: Callable) -> Callable:

        @wraps(func)
        def wrapper(*args, **kwargs) -> Optional[Any]:

            kwargs = fill_in_missing_kwargs(func, **kwargs)

            if self.origin_type is None and not self.mutual_exclusive:
                self.origin_type = get_origin_type_of_function_output(func)

            if len(self.parsed_output_type) == 0:
                self.parsed_output_type = self.parse_output_type(func, self.output_type, *args)
                self.parsed_output_type.append(self.origin_type)

            if self.model_name is None:
                self.initialize_rdr_model_name_and_load(func)

            func_output = {self.output_name: func(*args, **kwargs)}

            case, case_dict = create_case_from_method(func, func_output, *args, **kwargs)

            @self.fitting_decorator
            def fit():
                if self.expert is None:
                    self.expert = Human(answers_save_path=self.models_dir + f'/{self.model_name}/expert_answers')
                case_query = create_case_query_from_method(
                    func, func_output,
                    self.parsed_output_type,
                    self.mutual_exclusive,
                    args, kwargs,
                    case=case, case_dict=case_dict,
                    scenario=self.case_factory_metadata.scenario,
                    this_case_target_value=self.case_factory_metadata.this_case_target_value)
                output = self.rdr.fit_case(case_query, expert=self.expert,
                                           update_existing_rules=self.update_existing_rules)
                return output

            if self.fit and not self.use_generated_classifier and self.ask_now(case_dict):
                output = fit()
            else:
                if self.use_generated_classifier:
                    if self.generated_classifier is None:
                        model_path = os.path.join(self.models_dir, self.model_name)
                        self.generated_classifier = self.rdr.get_rdr_classifier_from_python_file(model_path)
                    output = self.generated_classifier(case)
                else:
                    output = self.rdr.classify(case)
                    if self.generate_dot_file:
                        eval_rule_tree = self.rdr.get_evaluated_rule_tree()
                        if not self._not_none_output_found or (eval_rule_tree and len(eval_rule_tree) > 1):
                            self.rdr.render_evaluated_rule_tree(self.models_dir + f'/{self.model_name}',
                                                                show_full_tree=True)
                        if eval_rule_tree and len(eval_rule_tree) > 1:
                            self._not_none_output_found = True

            if self.output_name in output:
                if self.origin_type == list:
                    return make_list(output[self.output_name])
                return output[self.output_name]
            else:
                return func_output[self.output_name]

        wrapper._rdr_decorator_instance = self

        return wrapper

    def initialize_rdr_model_name_and_load(self, func: Callable) -> None:
        self.model_name = get_func_rdr_model_name(func, include_file_name=True)
        self.load()

    def parse_output_type(self, func: Callable, output_type: Any, *args) -> List[Type]:
        parsed_output_type = []
        for ot in make_set(output_type):
            ot = get_arg_type_of_function_output(func, ot, *args, package_name=self.package_name)
            parsed_output_type.append(ot)
        return parsed_output_type

    def load(self):
        """
        Load the RDR model from the specified directory, otherwise create a new one.
        """
        self.rdr = GeneralRDR(save_dir=self.models_dir, model_name=self.model_name)


def fit_rdr_func(scenario: Callable, rdr_decorated_func: Callable, *func_args, **func_kwargs) -> None:
    rdr_decorated_func._rdr_decorator_instance.case_factory_metadata = CaseFactoryMetaData(scenario=scenario)
    rdr_decorated_func(*func_args, **func_kwargs)



from typing_extensions import Any, Callable, List, Optional, Tuple, Type, Union
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def animal_habitats_of_type_habitat(case: DataFrame) -> List[Habitat]:
    """Get possible value(s) for Animal.habitats  of type Habitat."""
    # Write your code here
    habitats = Habitat.land
    return habitats



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    is_a_mammal = case.species == Species.mammal
    return is_a_mammal



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type, Union
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def animal_habitats_of_type_habitat(case: DataFrame) -> List[Habitat]:
    """Get possible value(s) for Animal.habitats  of type Habitat."""
    # Write your code here
    habitats = Habitat.water
    return habitats



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    is_aquatic = case.aquatic == 1
    return is_aquatic



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type, Union
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def animal_habitats_of_type_habitat(case: DataFrame) -> List[Habitat]:
    """Get possible value(s) for Animal.habitats  of type Habitat."""
    # Write your code here
    habitats = {Habitat.land, Habitat.air}
    return habitats



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    is_a_bird = case.species == Species.bird
    has_legs = case.legs > 0
    return is_a_bird and has_legs



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type, Union
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def animal_habitats_of_type_habitat(case: DataFrame) -> List[Habitat]:
    """Get possible value(s) for Animal.habitats  of type Habitat."""
    # Write your code here
    habitats = Habitat.land
    return habitats



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    is_a_molusc = case.species == Species.molusc
    is_not_aquatic = case.aquatic == 0
    return is_a_molusc and is_not_aquatic



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type, Union
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def animal_habitats_of_type_habitat(case: DataFrame) -> List[Habitat]:
    """Get possible value(s) for Animal.habitats  of type Habitat."""
    # Write your code here
    habitats = {Habitat.land, Habitat.water}
    return habitats



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    is_a_molusc = case.species == Species.molusc
    is_aquatic = case.aquatic == 1
    has_legs = case.legs > 0
    return is_a_molusc and is_aquatic and has_legs



'===New Answer==='





from __future__ import annotations

import importlib
import inspect
import os
import sys
from functools import wraps
from types import ModuleType
from typing import Tuple, Callable, Dict, Any, Optional, Sequence, get_type_hints

from typing_extensions import Type, Optional, Callable, Any, Dict, TYPE_CHECKING, Union

from .datastructures.case import create_case, Case
from .datastructures.dataclasses import CaseQuery
from .utils import calculate_precision_and_recall, get_method_args_as_dict, get_func_rdr_model_name
from .utils import get_func_rdr_model_name, copy_case, make_set, update_case

if TYPE_CHECKING:
    from .rdr import RippleDownRules


def general_rdr_classify(classifiers_dict: Dict[str, Union[ModuleType, RippleDownRules]],
                         case: Any, modify_original_case: bool = False,
                         case_query: Optional[CaseQuery] = None) -> Dict[str, Any]:
    """
    Classify a case by going through all classifiers and adding the categories that are classified,
     and then restarting the classification until no more categories can be added.

    :param classifiers_dict: A dictionary mapping conclusion types to the classifiers that produce them.
    :param case: The case to classify.
    :param modify_original_case: Whether to modify the original case or create a copy and modify it.
    :param case_query: The case query to extract metadata from if needed.
    :return: The categories that the case belongs to.
    """
    conclusions = {}
    case = create_case(case)
    case_cp = copy_case(case) if not modify_original_case else case
    while True:
        new_conclusions = {}
        for attribute_name, rdr in classifiers_dict.items():
            pred_atts = rdr.classify(case_cp, case_query=case_query)
            if pred_atts is None and type(None) not in rdr.conclusion_type:
                continue
            if rdr.mutually_exclusive:
                if attribute_name not in conclusions or \
                        (attribute_name in conclusions and conclusions[attribute_name] != pred_atts):
                    conclusions[attribute_name] = pred_atts
                    new_conclusions[attribute_name] = pred_atts
            else:
                pred_atts = make_set(pred_atts)
                if attribute_name in conclusions:
                    pred_atts = {p for p in pred_atts if p not in conclusions[attribute_name]}
                if len(pred_atts) > 0:
                    new_conclusions[attribute_name] = pred_atts
                    if attribute_name not in conclusions:
                        conclusions[attribute_name] = set()
                    conclusions[attribute_name].update(pred_atts)
            if attribute_name in new_conclusions:
                temp_case_query = CaseQuery(case_cp, attribute_name, rdr.conclusion_type, rdr.mutually_exclusive)
                update_case(temp_case_query, new_conclusions)
        if len(new_conclusions) == 0 or len(classifiers_dict) == 1 and list(classifiers_dict.values())[
            0].mutually_exclusive:
            break
    return conclusions


def is_matching(classifier: Callable[[Any], Any], case_query: CaseQuery,
                pred_cat: Optional[Dict[str, Any]] = None) -> bool:
    """
    :param classifier: The RDR classifier to check the prediction of.
    :param case_query: The case query to check.
    :param pred_cat: The predicted category.
    :return: Whether the classifier prediction is matching case_query target or not.
    """
    if case_query.target is None:
        return False
    if pred_cat is None:
        pred_cat = classifier(case_query.case)
    if not isinstance(pred_cat, dict):
        pred_cat = {case_query.attribute_name: pred_cat}
    target = {case_query.attribute_name: case_query.target_value}
    precision, recall = calculate_precision_and_recall(pred_cat, target)
    return all(recall) and all(precision)


def load_or_create_func_rdr_model(func, model_dir: str, rdr_type: Type[RippleDownRules],
                                  **rdr_kwargs) -> RippleDownRules:
    """
    Load the RDR model of the function if it exists, otherwise create a new one.

    :param func: The function to load the model for.
    :param model_dir: The directory where the model is stored.
    :param rdr_type: The type of the RDR model to load.
    :param rdr_kwargs: Additional arguments to pass to the RDR constructor in the case of a new model.
    """
    model_name = get_func_rdr_model_name(func)
    model_path = os.path.join(model_dir, model_name, f"{model_name}.py")
    if os.path.exists(model_path):
        rdr = rdr_type.load(load_dir=model_dir, model_name=model_name)
    else:
        rdr = rdr_type(**rdr_kwargs)
    return rdr


def get_an_updated_case_copy(case: Case, conclusion: Callable, attribute_name: str, conclusion_type: Tuple[Type, ...],
                             mutually_exclusive: bool) -> Case:
    """
    :param case: The case to copy and update.
    :param conclusion: The conclusion to add to the case.
    :param attribute_name: The name of the attribute to update.
    :param conclusion_type: The type of the conclusion to update.
    :param mutually_exclusive: Whether the rule belongs to a mutually exclusive RDR.
    :return: A copy of the case updated with the given conclusion.
    """
    case_cp = copy_case(case)
    temp_case_query = CaseQuery(case_cp, attribute_name, conclusion_type,
                                mutually_exclusive=mutually_exclusive)
    output = conclusion(case_cp)
    if not isinstance(output, Dict):
        output = {attribute_name: output}
    update_case(temp_case_query, output)
    return case_cp

def enable_gui():
    """
    Enable the GUI for Ripple Down Rules if available.
    """
    try:
        from .user_interface.gui import RDRCaseViewer
        viewer = RDRCaseViewer()
    except ImportError:
        pass


def create_case_from_method(func: Callable,
                            func_output: Dict[str, Any],
                            *args, **kwargs) -> Tuple[Case, Dict[str, Any]]:
    """
    Create a Case from the function and its arguments.

    :param func: The function to create a case from.
    :param func_output: A dictionary containing the output of the function, where the key is the output name.
    :param args: The positional arguments of the function.
    :param kwargs: The keyword arguments of the function.
    :return: A Case object representing the case.
    """
    case_dict = get_method_args_as_dict(func, *args, **kwargs)
    case_dict.update(func_output)
    case_name = get_func_rdr_model_name(func)
    return Case(dict, id(case_dict), case_name, case_dict, **case_dict), case_dict


class MockRDRDecorator:
    def __init__(self, models_dir: str):
        self.models_dir = models_dir
    def decorator(self, func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Optional[Any]:
            model_dir = get_func_rdr_model_name(func, include_file_name=True)
            model_name = get_func_rdr_model_name(func, include_file_name=False)
            rdr = importlib.import_module(os.path.join(self.models_dir, model_dir, f"{model_name}_rdr.py"))
            func_output = {"output_": func(*args, **kwargs)}
            case, case_dict = create_case_from_method(func, func_output, *args, **kwargs)
            return rdr.classify(case)
        return wrapper


def create_case_query_from_method(func: Callable,
                                  func_output: Dict[str, Any],
                                  output_type: Sequence[Type],
                                  mutual_exclusive: bool,
                                  func_args: Tuple[Any, ...], func_kwargs: Dict[str, Any],
                                  case: Optional[Case] = None,
                                  case_dict: Optional[Dict[str, Any]] = None,
                                  scenario: Optional[Callable] = None,
                                  this_case_target_value: Optional[Any] = None,) -> CaseQuery:
    """
    Create a CaseQuery from the function and its arguments.

    :param func: The function to create a case from.
    :param func_output: The output of the function as a dictionary, where the key is the output name.
    :param output_type: The type of the output as a sequence of types.
    :param mutual_exclusive: If True, the output types are mutually exclusive.
    :param func_args: The positional arguments of the function.
    :param func_kwargs: The keyword arguments of the function.
    :param case: The case to create.
    :param case_dict: The dictionary of the case.
    :param scenario: The scenario that produced the given case.
    :param this_case_target_value: The target value for the case.
    :return: A CaseQuery object representing the case.
    """
    output_type = make_set(output_type)
    if case is None or case_dict is None:
        case, case_dict = create_case_from_method(func, func_output, *func_args, **func_kwargs)
    scope = func.__globals__
    scope.update(case_dict)
    try:
        func_args_type_hints = get_type_hints(func)
    except NameError:
        # use inspect to get the type hints if get_type_hints fails
        func_args_type_hints = {k: v.annotation for k, v in inspect.signature(func).parameters.items()
                                  if v.annotation is not inspect._empty}
        # add return type hint from the function signature
        return_annotation = inspect.signature(func).return_annotation
        if return_annotation is not inspect._empty:
            func_args_type_hints['return'] = return_annotation
        types_dict = {t.__name__: t for t in output_type}
        scope.update(types_dict)
        types_names = list(types_dict.keys())
        for k, v in func_args_type_hints.items():
            for t_name in types_names:
                if isinstance(v, str) and t_name in v:
                    func_args_type_hints[k] = eval(v, scope)
    output_name = list(func_output.keys())[0]
    func_args_type_hints.update({output_name: Union[tuple(output_type)]})
    return CaseQuery(case, output_name, tuple(output_type),
                     mutual_exclusive, scope=scope, scenario=scenario, this_case_target_value=this_case_target_value,
                     is_function=True, function_args_type_hints=func_args_type_hints)



from __future__ import annotations
from dataclasses import dataclass, field
from typing_extensions import List


@dataclass(unsafe_hash=True)
class PhysicalObject:
    """
    A physical object is an object that can be contained in a container.
    """
    name: str
    contained_objects: List[PhysicalObject] = field(default_factory=list, hash=False)


@dataclass(unsafe_hash=True)
class Part(PhysicalObject):
    ...


@dataclass(unsafe_hash=True)
class Robot(PhysicalObject):
    parts: List[Part] = field(default_factory=list, hash=False)


def my_robot_factory() -> Robot:
    """
    Factory function to create a simple robot with parts and containment relationships.
    """
    part_a = Part(name="A")
    part_b = Part(name="B")
    part_c = Part(name="C")
    robot = Robot("pr2", parts=[part_a])
    part_a.contained_objects = [part_b]
    part_b.contained_objects = [part_c]

    return robot



from ripple_down_rules.utils import make_set
from ripple_down_rules.helpers import get_an_updated_case_copy
from typing_extensions import Optional, Set
from ripple_down_rules.datastructures.case import Case, create_case
from .robot_contained_objects_mcrdr_defs import *


attribute_name = 'contained_objects'
conclusion_type = (set, PhysicalObject, list,)
mutually_exclusive = False
name = 'contained_objects'
case_type = Robot
case_name = 'Robot'


def classify(case: Robot, **kwargs) -> Set[PhysicalObject]:
    if not isinstance(case, Case):
        case = create_case(case, max_recursion_idx=3)
    conclusions = set()

    if conditions_298609776593271728826836208156881692889(case):
        conclusions.update(make_set(conclusion_298609776593271728826836208156881692889(case)))
    return conclusions



from __future__ import annotations

import inspect
import uuid
from collections import defaultdict
from dataclasses import dataclass, field, Field, fields
from enum import Enum
from functools import lru_cache

import pydot
import rustworkx as rx
from typing_extensions import Any, Type, ClassVar, Dict, List, Optional, Tuple

from .field_info import FieldInfo
from .. import logger
from ..rules import Rule
from ..utils import recursive_subclasses


class Direction(Enum):
    OUTBOUND = False
    INBOUND = True


class Relation(str, Enum):
    has = "has"
    isA = "isA"
    dependsOn = "dependsOn"


@dataclass(unsafe_hash=True)
class TrackedObjectMixin:
    """
    A class that is used as a base class to all classes that needs to be tracked for RDR inference, and reasoning.
    """
    _rdr_tracked_object_id: int = field(init=False, repr=False, hash=False,
                                        compare=False, default_factory=lambda: uuid.uuid4().int)
    """
    The unique identifier of the conclusion.
    """
    _dependency_graph: ClassVar[rx.PyDAG[Type[TrackedObjectMixin]]] = rx.PyDAG()
    """
    A graph that represents the relationships between all tracked objects.
    """
    _class_graph_indices: ClassVar[Dict[Type[TrackedObjectMixin], int]] = {}
    """
    The index of the current class in the dependency graph.
    """
    _edges: ClassVar[Dict[Relation, List[Tuple[int, int]]]] = defaultdict(list)
    """
    All the edges indexed by relation type.
    """
    _overridden_by: Type[TrackedObjectMixin] = field(init=False, repr=False, hash=False,
                                                     compare=False, default=None)
    """
    Whether the class has been overridden by a subclass.
    This is used to only include the new class in the dependency graph, not the overridden class.
    """

    @classmethod
    def _reset_dependency_graph(cls) -> None:
        """
        Reset the dependency graph and all class indices.
        """
        cls._dependency_graph = rx.PyDAG()
        cls._class_graph_indices = {}
        cls._edges = defaultdict(list)
        cls.parse_fields.cache_clear()
        cls.get_fields.cache_clear()
        cls.parse_field.cache_clear()

    @classmethod
    def _my_graph_idx(cls):
        return cls._class_graph_indices[cls]

    @classmethod
    def make_class_dependency_graph(cls, composition: bool = True):
        """
        Create a direct acyclic graph containing the class hierarchy.

        :param composition: If True, the class dependency graph will include composition relations.
        """
        classes_to_track = recursive_subclasses(cls) + [Rule] + recursive_subclasses(Rule)
        for clazz in classes_to_track:
            cls._add_class_to_dependency_graph(clazz)

        for clazz in cls._class_graph_indices:
            bases = [base for base in clazz.__bases__ if
                     base.__module__ not in ["builtins"] and base in cls._class_graph_indices]

            for base in bases:
                cls._add_class_to_dependency_graph(base)
                clazz_idx = cls._class_graph_indices[clazz]
                base_idx = cls._class_graph_indices[base]
                if (clazz_idx, base_idx) in cls._edges[Relation.isA] or base._overridden_by == clazz:
                    continue
                cls.add_edge(clazz_idx, base_idx, Relation.isA)

        if not composition:
            return
        for clazz, idx in cls._class_graph_indices.items():
            if clazz.__module__ == "builtins":
                continue
            TrackedObjectMixin.parse_fields(clazz)

    @staticmethod
    @lru_cache(maxsize=None)
    def parse_fields(clazz) -> None:
        for f in TrackedObjectMixin.get_fields(clazz):

            logger.debug("=" * 80)
            logger.debug(f"Processing Field {clazz.__name__}.{f.name}: {f.type}.")

            # skip private fields
            if f.name.startswith("_"):
                logger.debug(f"Skipping since the field starts with _.")
                continue

            field_info = FieldInfo(clazz, f)
            TrackedObjectMixin.parse_field(field_info)

    @staticmethod
    @lru_cache(maxsize=None)
    def get_fields(clazz) -> List[Field]:
        skip_fields = []
        bases = [base for base in clazz.__bases__ if issubclass(base, TrackedObjectMixin)]
        for base in bases:
            skip_fields.extend(TrackedObjectMixin.get_fields(base))

        result = [cls_field for cls_field in fields(clazz) if cls_field not in skip_fields]

        return result

    @staticmethod
    @lru_cache(maxsize=None)
    def parse_field(field_info: FieldInfo):
        parent_idx = TrackedObjectMixin._class_graph_indices[field_info.clazz]
        field_cls: Optional[Type[TrackedObjectMixin]] = None
        field_relation = Relation.has
        if len(field_info.type) == 1 and issubclass(field_info.type[0], TrackedObjectMixin):
            field_cls = field_info.type[0]
        else:
            # TODO: Create a new TrackedObjectMixin class for new type
            logger.debug(f"Skipping unhandled field type: {field_info.type}")
            # logger.debug(f"Creating new TrackedObject type for builtin type {field_info.type}.")
            # field_cls = cls._create_tracked_object_class_for_field(field_info)

        if field_cls is not None:
            field_cls_idx = TrackedObjectMixin._class_graph_indices.get(field_cls, None)
            if field_cls_idx is not None and (parent_idx, field_cls_idx) in TrackedObjectMixin._edges[Relation.has]:
                return
            elif field_cls_idx is None:
                TrackedObjectMixin._add_class_to_dependency_graph(field_cls)
                field_cls_idx = TrackedObjectMixin._class_graph_indices[field_cls]
            TrackedObjectMixin.add_edge(parent_idx, field_cls_idx, field_relation)

    @classmethod
    def add_edge(cls, parent_idx: int, child_idx: int, relation: Relation):
        cls._dependency_graph.add_edge(parent_idx, child_idx, relation)
        cls._edges[relation].append((parent_idx, child_idx))

    @classmethod
    def _create_tracked_object_class_for_field(cls, field_info: FieldInfo):
        raise NotImplementedError

    @classmethod
    def to_dot(cls, filepath: str, format='png') -> None:
        if not filepath.endswith(f".{format}"):
            filepath += f".{format}"
        dot_str = cls._dependency_graph.to_dot(
            lambda node: dict(
                color='black', fillcolor='lightblue', style='filled', label=node.__name__),
            lambda edge: dict(color='black', style='solid', label=edge))
        dot = pydot.graph_from_dot_data(dot_str)[0]
        dot.write(filepath, format=format)

    @classmethod
    def _add_class_to_dependency_graph(cls, class_to_add: Type[TrackedObjectMixin]) -> None:
        """
        Add a class to the dependency graph.
        """
        if class_to_add not in cls._class_graph_indices:
            if not issubclass(class_to_add, TrackedObjectMixin):
                class_to_add._overridden_by = None
            cls_idx = cls._dependency_graph.add_node(class_to_add._overridden_by or class_to_add)
            cls._class_graph_indices[class_to_add] = cls_idx
            if class_to_add._overridden_by:
                cls._class_graph_indices[class_to_add._overridden_by] = cls_idx

    def __getattribute__(self, name: str) -> Any:
        # if name not in [f.name for f in fields(TrackedObjectMixin)] + ['has', 'is_a', 'depends_on']\
        #         and not name.startswith("_"):
        #     self._record_dependency(name)
        return object.__getattribute__(self, name)

    def _record_dependency(self, attr_name):
        # Inspect stack to find instance of CallableExpression
        for frame_info in inspect.stack():
            func_name = frame_info.function
            local_self = frame_info.frame.f_locals.get("self", None)
            if (
                    func_name == "__call__" and
                    local_self is not None and
                    type(local_self).__module__ == "callable_expression" and
                    type(local_self).__name__ == "CallableExpression"
            ):
                logger.debug("TrackedObject used inside CallableExpression")
                break


annotations = TrackedObjectMixin.__annotations__
for val in [f.name for f in fields(TrackedObjectMixin)]:
    annotations.pop(val, None)

X = TrackedObjectMixin



from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.milk == 1



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.aquatic == 1



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.feathers == 1



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.backbone == 0 and case.breathes == 0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.fins == 0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.feathers == 1



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.eggs == 1 and case.backbone == 0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.backbone == 1



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.backbone == 1 and case.tail == 1



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.backbone == 0 and case.eggs == 0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.breathes == 0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, MappedAnimal, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat

def conditions_for_mapped_animal_species_of_type_species(case: MappedAnimal) -> bool:
    """Get conditions on whether it's possible to conclude a value for MappedAnimal.species  of type Species."""
    # Write your code here
    return case.legs == 0



'===New Answer==='





from __future__ import annotations

import ast
import importlib
import json
import os
import sys
from abc import ABC, abstractmethod
from copy import copy
from dataclasses import is_dataclass
from io import TextIOWrapper
from os.path import dirname
from pathlib import Path
from types import NoneType, ModuleType

from ripple_down_rules.datastructures.dataclasses import CaseFactoryMetaData
from . import logger
from .failures import RDRLoadError

try:
    from matplotlib import pyplot as plt
    Figure = plt.Figure
except ImportError as e:
    logger.debug(f"{e}: matplotlib is not installed")
    matplotlib = None
    Figure = None
    plt = None

from sqlalchemy.orm import DeclarativeBase as SQLTable
from typing_extensions import List, Optional, Dict, Type, Union, Any, Self, Tuple, Callable, Set

from .datastructures.callable_expression import CallableExpression
from .datastructures.case import Case, CaseAttribute, create_case
from .datastructures.dataclasses import CaseQuery
from .datastructures.enums import MCRDRMode, RDREdge
from .experts import Expert, Human
from .helpers import is_matching, general_rdr_classify, get_an_updated_case_copy
from .rules import Rule, SingleClassRule, MultiClassTopRule, MultiClassStopRule, MultiClassRefinementRule, \
    MultiClassFilterRule

try:
    from .user_interface.gui import RDRCaseViewer
except ImportError as e:
    RDRCaseViewer = None
from .utils import draw_tree, make_set, SubclassJSONSerializer, make_list, get_type_from_string, \
    is_value_conflicting, extract_function_source, extract_imports, get_full_class_name, \
    is_iterable, str_to_snake_case, get_import_path_from_path, get_imports_from_types, render_tree, \
    get_types_to_import_from_func_type_hints, get_function_return_type, get_file_that_ends_with, \
    get_and_import_python_module, get_and_import_python_modules_in_a_package, get_type_from_type_hint, \
    are_results_subclass_of_types


class RippleDownRules(SubclassJSONSerializer, ABC):
    """
    The abstract base class for the ripple down rules classifiers.
    """
    fig: Optional[Figure] = None
    """
    The figure to draw the tree on.
    """
    expert_accepted_conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the expert has accepted, such that they are not asked again.
    """
    _generated_python_file_name: Optional[str] = None
    """
    The name of the generated python file.
    """
    name: Optional[str] = None
    """
    The name of the classifier.
    """
    case_type: Optional[Type] = None
    """
    The type of the case (input) to the RDR classifier.
    """
    case_name: Optional[str] = None
    """
    The name of the case type.
    """
    metadata_folder: str = "rdr_metadata"
    """
    The folder to save the metadata of the RDR classifier.
    """
    model_name: Optional[str] = None
    """
    The name of the model. If None, the model name will be the generated python file name.
    """
    mutually_exclusive: Optional[bool] = None
    """
    Whether the output of the classification of this rdr allows only one possible conclusion or not.
    """

    def __init__(self, start_rule: Optional[Rule] = None,
                 save_dir: Optional[str] = None, model_name: Optional[str] = None):
        """
        :param start_rule: The starting rule for the classifier.
        :param save_dir: The directory to save the classifier to.
        """
        self.model_name: Optional[str] = model_name
        self.save_dir: Optional[str] = save_dir
        self.start_rule: Optional[Rule] = start_rule
        self.fig: Optional[Figure] = None
        self.viewer: Optional[RDRCaseViewer] = RDRCaseViewer.instances[0]\
            if RDRCaseViewer and any(RDRCaseViewer.instances) else None
        self.input_node: Optional[Rule] = None
        self.update_model()

    def update_model(self):
        """
        Update the model from the model directory if it exists.
        """
        if self.save_dir is not None and self.model_name is not None:
            model_path = os.path.join(self.save_dir, self.model_name)
            if os.path.exists(model_path):
                try:
                    self.update_from_python(model_path, update_rule_tree=True)
                except (FileNotFoundError, ModuleNotFoundError) as e:
                    pass

    def write_rdr_metadata_to_pyton_file(self, file: TextIOWrapper):
        """
        Write the metadata of the RDR classifier to a python file.

        :param file: The file to write the metadata to.
        """
        file.write(f"name = \'{self.name}\'\n")
        file.write(f"case_type = {self.case_type.__name__ if self.case_type is not None else None}\n")
        file.write(f"case_name = \'{self.case_name}\'\n")

    def update_rdr_metadata_from_python(self, module: ModuleType):
        """
        Update the RDR metadata from the module that contains the RDR classifier function.

        :param module: The module that contains the RDR classifier function.
        """
        try:
            self.name = module.name if hasattr(module, "name") else self.start_rule.conclusion_name
            self.case_type = module.case_type
            self.case_name = module.case_name if hasattr(module, "case_name") else f"{self.case_type.__name__}.{self.name}"
        except AttributeError as e:
            logger.warning(f"Could not update the RDR metadata from the module {module.__name__}. "
                           f"Make sure the module has the required attributes: {e}")

    def render_evaluated_rule_tree(self, filename: str, show_full_tree: bool = False) -> None:
        if show_full_tree:
            start_rule = self.start_rule if self.input_node is None else self.input_node
            render_tree(start_rule, use_dot_exporter=True, filename=filename)
        else:
            evaluated_rules = self.get_evaluated_rule_tree()
            if evaluated_rules is not None and len(evaluated_rules) > 0:
                render_tree(evaluated_rules[0], use_dot_exporter=True, filename=filename,
                            only_nodes=evaluated_rules)

    def get_contributing_rules(self) -> Optional[List[Rule]]:
        """
        Get the contributing rules of the classifier.

        :return: The contributing rules.
        """
        if self.start_rule is None:
            return None
        return [r for r in self.get_fired_rule_tree() if r.contributed]

    def get_fired_rule_tree(self) -> Optional[List[Rule]]:
        """
        Get the fired rule tree of the classifier.

        :return: The fired rule tree.
        """
        if self.start_rule is None:
            return None
        return [r for r in self.get_evaluated_rule_tree() if r.fired]

    def get_evaluated_rule_tree(self) -> Optional[List[Rule]]:
        """
        Get the evaluated rule tree of the classifier.

        :return: The evaluated rule tree.
        """
        if self.start_rule is None:
            return None
        start_rule = self.start_rule
        evaluated_rule_tree = [r for r in [start_rule] + list(start_rule.descendants) if r.evaluated]
        return evaluated_rule_tree

    def save(self, save_dir: Optional[str] = None, model_name: Optional[str] = None,
             package_name: Optional[str] = None) -> str:
        """
        Save the classifier to a file.

        :param save_dir: The directory to save the classifier to.
        :param model_name: The name of the model to save. If None, a default name is generated.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :return: The name of the saved model.
        """
        save_dir = save_dir or self.save_dir
        if save_dir is None:
            raise ValueError("The save directory cannot be None. Please provide a valid directory to save"
                             " the classifier.")
        if not os.path.exists(save_dir + '/__init__.py'):
            os.makedirs(save_dir, exist_ok=True)
            with open(save_dir + '/__init__.py', 'w') as f:
                f.write("from . import *\n")
        if model_name is not None:
            self.model_name = model_name
        elif self.model_name is None:
            self.model_name = self.generated_python_file_name
        model_dir = os.path.join(save_dir, self.model_name)
        os.makedirs(model_dir, exist_ok=True)
        json_dir = os.path.join(model_dir, self.metadata_folder)
        os.makedirs(json_dir, exist_ok=True)
        self.to_json_file(os.path.join(json_dir, self.model_name))
        self._write_to_python(model_dir, package_name=package_name)
        return self.model_name

    @classmethod
    def load(cls, load_dir: str, model_name: str,
             package_name: Optional[str] = None) -> Self:
        """
        Load the classifier from a file.

        :param load_dir: The path to the model directory to load the classifier from.
        :param model_name: The name of the model to load.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        rdr: Optional[RippleDownRules] = None
        model_dir = os.path.join(load_dir, model_name)
        json_file = os.path.join(model_dir, cls.metadata_folder, model_name)
        if os.path.exists(json_file + ".json"):
            rdr = cls.from_json_file(json_file)
        try:
            if rdr is None:
                rdr = cls.from_python(model_dir, parent_package_name=package_name)
            else:
                rdr.update_from_python(model_dir, parent_package_name=package_name)
            rdr.to_json_file(json_file)
        except (FileNotFoundError, ValueError, SyntaxError, ModuleNotFoundError) as e:
            logger.warning(f"Could not load the python file for the model {model_name} from {model_dir}. "
                           f"Make sure the file exists and is valid.")
            if rdr is None:
                raise RDRLoadError(f"Could not load the rdr model {model_name} from {model_dir}, error is {e}")
            rdr.save(save_dir=load_dir, model_name=model_name, package_name=package_name)
        rdr.save_dir = load_dir
        rdr.model_name = model_name
        return rdr

    @classmethod
    def from_python(cls, model_path: str,
                    python_file_path: Optional[str] = None,
                    parent_package_name: Optional[str] = None) -> Self:
        """
        Load the RDR classifier from a generated python file.

        :param model_path: The directory where the generated python file is located.
        :param python_file_path: The path to the generated python file that contains the RDR classifier function.
        :param parent_package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :return: An instance of the RDR classifier.
        """
        rdr = cls()
        rdr.update_from_python(model_path, parent_package_name=parent_package_name, python_file_path=python_file_path,
                               update_rule_tree=True)
        return rdr

    @abstractmethod
    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None):
        """
        Write the tree of rules as source code to a file.

        :param model_dir: The path to the directory to write the source code to.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        pass

    def fit(self, case_queries: List[CaseQuery],
            expert: Optional[Expert] = None,
            n_iter: int = None,
            animate_tree: bool = False,
            **kwargs_for_fit_case):
        """
        Fit the classifier to a batch of cases and categories.

        :param case_queries: The cases and categories to fit the classifier to.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param n_iter: The number of iterations to fit the classifier for.
        :param animate_tree: Whether to draw the tree while fitting the classifier.
        :param kwargs_for_fit_case: The keyword arguments to pass to the fit_case method.
        """
        targets = []
        if animate_tree:
            if plt is None:
                raise ImportError("matplotlib is not installed, cannot animate the tree.")
            plt.ion()
        i = 0
        stop_iterating = False
        num_rules: int = 0
        while not stop_iterating:
            for case_query in case_queries:
                pred_cat = self.fit_case(case_query, expert=expert, clear_expert_answers=False,
                                         **kwargs_for_fit_case)
                if case_query.target is None:
                    continue
                target = {case_query.attribute_name: case_query.target(case_query.case)}
                if len(targets) < len(case_queries):
                    targets.append(target)
                match = is_matching(self.classify, case_query, pred_cat)
                if not match:
                    print(f"Predicted: {pred_cat} but expected: {target}")
                if animate_tree and len(self.start_rule.descendants) > num_rules:
                    num_rules = len(self.start_rule.descendants)
                    self.update_figures()
            i += 1
            all_predictions = [1 if is_matching(self.classify, case_query) else 0 for case_query in case_queries
                               if case_query.target is not None]
            all_pred = sum(all_predictions)
            logger.info(f"Accuracy: {all_pred}/{len(targets)}")
            all_predicted = targets and all_pred == len(targets)
            num_iter_reached = n_iter and i >= n_iter
            stop_iterating = all_predicted or num_iter_reached
            if stop_iterating:
                break
        logger.info(f"Finished training in {i} iterations")
        if animate_tree:
            plt.ioff()
            plt.show()

    def __call__(self, case: Union[Case, SQLTable]) -> Union[CallableExpression, Dict[str, CallableExpression]]:
        return self.classify(case)

    def classify(self, case: Union[Case, SQLTable], modify_case: bool = False, case_query: Optional[CaseQuery] = None) \
            -> Optional[Union[CallableExpression, Dict[str, CallableExpression]]]:
        """
        Classify a case using the RDR classifier.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case attributes with the conclusion or not.
        :param case_query: The case query containing the case to classify and the target category to compare the case with.
        :return: The category that the case belongs to.
        """
        if self.start_rule is not None:
            for rule in [self.start_rule] + list(self.start_rule.descendants):
                rule.reset()
        if self.start_rule is not None and self.start_rule.parent is None:
            if self.input_node is None:
                self.input_node = type(self.start_rule)(_parent=None, uid='0')
                self.input_node.evaluated = False
                self.input_node.fired = False
            self.start_rule.parent = self.input_node
            self.start_rule.weight = RDREdge.Empty
        if self.input_node is not None:
            data = case.__dict__ if is_dataclass(case) else case
            if hasattr(case, "items"):
                self.input_node.name = json.dumps({k: str(v) for k, v in data.items()}, indent=4)
            else:
                self.input_node.name = str(data)
        return self._classify(case, modify_case=modify_case, case_query=case_query)

    @abstractmethod
    def _classify(self, case: Union[Case, SQLTable], modify_case: bool = False,
                  case_query: Optional[CaseQuery] = None) \
            -> Optional[Union[CallableExpression, Dict[str, CallableExpression]]]:
        """
        Classify a case.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case attributes with the conclusion or not.
        :param case_query: The case query containing the case to classify and the target category to compare the case with.
        :return: The category that the case belongs to.
        """
        pass

    def fit_case(self, case_query: CaseQuery,
                 expert: Optional[Expert] = None,
                 update_existing_rules: bool = True,
                 scenario: Optional[Callable] = None,
                 ask_now: Callable = lambda _: True,
                 clear_expert_answers: bool = True,
                 **kwargs) \
            -> Union[CallableExpression, Dict[str, CallableExpression]]:
        """
        Fit the classifier to a case and ask the expert for refinements or alternatives if the classification is
        incorrect by comparing the case with the target category.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param update_existing_rules: Whether to update the existing same conclusion type rules that already gave
        some conclusions with the type required by the case query.
        :param scenario: The scenario at which the case was created, this is used to recreate the case if needed.
        :param ask_now: Whether to ask the expert for refinements or alternatives.
        :param clear_expert_answers: Whether to clear expert answers after saving the new rule.
        :return: The category that the case belongs to.
        """
        if case_query is None:
            raise ValueError("The case query cannot be None.")

        self.name = case_query.attribute_name if self.name is None else self.name
        self.case_type = case_query.case_type if self.case_type is None else self.case_type
        self.case_name = case_query.case_name if self.case_name is None else self.case_name
        case_query.scenario = scenario if case_query.scenario is None else case_query.scenario
        case_query.rdr = self

        expert = expert or Human(answers_save_path=self.save_dir + '/expert_answers'
                                 if self.save_dir else None)
        if case_query.target is None:
            case_query_cp = copy(case_query)
            conclusions = self.classify(case_query_cp.case, modify_case=True, case_query=case_query_cp)
            if (self.should_i_ask_the_expert_for_a_target(conclusions, case_query_cp, update_existing_rules)
                    and ask_now(case_query_cp.case)):
                expert.ask_for_conclusion(case_query_cp)
                case_query.target = case_query_cp.target
            if case_query.target is None:
                return self.classify(case_query.case)

        self.update_start_rule(case_query, expert)

        fit_case_result = self._fit_case(case_query, expert=expert, **kwargs)

        if self.save_dir is not None:
            self.save()
            if clear_expert_answers:
                expert.clear_answers()

        return fit_case_result

    @staticmethod
    def should_i_ask_the_expert_for_a_target(conclusions: Union[Any, Dict[str, Any]],
                                             case_query: CaseQuery,
                                             update_existing: bool) -> bool:
        """
        Determine if the rdr should ask the expert for the target of a given case query.

        :param conclusions: The conclusions of the case.
        :param case_query: The query containing the case to classify.
        :param update_existing: Whether to update rules that gave the required type of conclusions.
        :return: True if the rdr should ask the expert, False otherwise.
        """
        if conclusions is None and type(None) not in case_query.core_attribute_type:
            return True
        elif is_iterable(conclusions) and len(conclusions) == 0:
            return True
        elif isinstance(conclusions, dict):
            if case_query.attribute_name not in conclusions:
                return True
            conclusions = conclusions[case_query.attribute_name]
        conclusion_types = map(type, make_list(conclusions))
        if not any(ct in case_query.core_attribute_type for ct in conclusion_types):
            return True
        elif update_existing:
            return True
        else:
            return False

    @abstractmethod
    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CallableExpression, Dict[str, CallableExpression]]:
        """
        Fit the RDR on a case, and ask the expert for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        pass

    @abstractmethod
    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        pass

    def update_figures(self):
        """
        Update the figures of the classifier.
        """
        if isinstance(self, GeneralRDR):
            for i, (rdr_name, rdr) in enumerate(self.start_rules_dict.items()):
                if not rdr.fig:
                    rdr.fig = plt.figure(f"Rule {i}: {rdr_name}")
                draw_tree(rdr.start_rule, rdr.fig)
        else:
            if not self.fig:
                self.fig = plt.figure(0)
            draw_tree(self.start_rule, self.fig)

    @property
    def type_(self):
        return self.__class__

    @classmethod
    def get_json_file_path(cls, model_path: str) -> str:
        """
        Get the path to the saved json file.

        :param model_path : The path to the model directory.
        :return: The path to the saved model.
        """
        model_name = cls.get_model_name_from_model_path(model_path)
        return os.path.join(model_path, cls.metadata_folder, f"{model_name}.json")

    @classmethod
    def get_generated_cases_file_path(cls, model_path: str) -> str:
        """
        Get the path to the python file that contains the RDR classifier cases.

        :param model_path : The path to the model directory.
        :return: The path to the generated python file.
        """
        return cls.get_generated_python_file_path(model_path).replace(".py", "_cases.py")

    @classmethod
    def get_generated_defs_file_path(cls, model_path: str) -> str:
        """
        Get the path to the python file that contains the RDR classifier function definitions.

        :param model_path : The path to the model directory.
        :return: The path to the generated python file.
        """
        return cls.get_generated_python_file_path(model_path).replace(".py", "_defs.py")

    @classmethod
    def get_generated_python_file_path(cls, model_path: str) -> str:
        """
        Get the path to the python file that contains the RDR classifier function.

        :param model_path : The path to the model directory.
        :return: The path to the generated python file.
        """
        model_name = cls.get_model_name_from_model_path(model_path)
        return os.path.join(model_path, f"{model_name}.py")

    @classmethod
    def get_model_name_from_model_path(cls, model_path: str) -> str:
        """
        Get the model name from the model path.

        :param model_path: The path to the model directory.
        :return: The name of the model.
        """
        file_name = get_file_that_ends_with(model_path, f"_{cls.get_acronym().lower()}.py")
        if file_name is None:
            raise FileNotFoundError(f"Could not find the python file for the model in the given path: {model_path}.")
        return file_name.replace('.py', '')

    @property
    def generated_python_file_name(self) -> str:
        if self._generated_python_file_name is None:
            self._generated_python_file_name = self._default_generated_python_file_name
        return self._generated_python_file_name

    @generated_python_file_name.setter
    def generated_python_file_name(self, value: str):
        """
        Set the generated python file name.
        :param value: The new value for the generated python file name.
        """
        self._generated_python_file_name = value

    @property
    @abstractmethod
    def _default_generated_python_file_name(self) -> str:
        """
        :return: The default generated python file name.
        """
        pass

    @abstractmethod
    def update_from_python(self, model_dir: str, parent_package_name: Optional[str] = None,
                           python_file_path: Optional[str] = None,
                           update_rule_tree: bool = False):
        """
        Update the rules from the generated python file, that might have been modified by the user.

        :param model_dir: The directory where the generated python file is located.
        :param parent_package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :param python_file_path: The path to the generated python file that contains the RDR classifier function.
        :param update_rule_tree: Whether to update the rule tree from the python file or not.
        """
        pass

    @classmethod
    def get_acronym(cls) -> str:
        """
        :return: The acronym of the classifier.
        """
        if cls.__name__ == "GeneralRDR":
            return "RDR"
        elif cls.__name__ == "MultiClassRDR":
            return "MCRDR"
        else:
            return "SCRDR"

    def get_rdr_classifier_from_python_file(self, package_name: str) -> Callable[[Any], Any]:
        """
        :param package_name: The name of the package that contains the RDR classifier function.
        :return: The module that contains the rdr classifier function.
        """
        # remove from imports if exists first
        package_name = get_import_path_from_path(package_name)
        name = f"{package_name}.{self.generated_python_file_name}" if package_name else self.generated_python_file_name
        module = importlib.import_module(name)
        importlib.reload(module)
        return module.classify


class TreeBuilder(ast.NodeVisitor, ABC):
    """Parses an AST of nested if-elif statements and reconstructs the tree."""

    def __init__(self):
        self.root: Optional[Rule] = None
        self.current_parent: Optional[Rule] = None
        self.current_edge: Optional[RDREdge] = None
        self.default_conclusion: Optional[str] = None

    def visit_FunctionDef(self, node):
        """Finds the main function and starts parsing its body."""
        for stmt in node.body:
            self.visit(stmt)

    def visit_If(self, node):
        """Handles if-elif blocks and creates nodes."""
        condition = self.get_condition_name(node.test)
        if condition is None:
            return
        rule_uid = condition.split("conditions_")[1]

        new_rule_type = self.get_new_rule_type(node)
        new_node = new_rule_type(conditions=condition, _parent=self.current_parent, uid=rule_uid)
        if self.current_parent is not None:
            self.update_current_parent(new_node)

        if self.current_parent is None and self.root is None:
            self.root = new_node

        self.current_parent = new_node

        # Parse the body of the if statement
        for stmt in node.body:
            self.current_edge = self.get_refinement_edge(node)
            self.current_parent = new_node
            self.visit(stmt)

        # Parse elif/else
        for stmt in node.orelse:
            self.current_edge = self.get_alternative_edge(node)
            self.current_parent = new_node
            if isinstance(stmt, ast.If): # elif case
                self.visit_If(stmt)
            else:  # else case (return)
                self.process_else_statement(stmt)
        self.current_parent = new_node
        self.current_edge = None

    @abstractmethod
    def process_else_statement(self, stmt: ast.AST):
        """
        Process the else statement in the if-elif-else block.

        :param stmt: The else statement to process.
        """
        pass

    @abstractmethod
    def get_refinement_edge(self, node: ast.AST) -> RDREdge:
        """
        :param node: The current AST node to determine the edge type from.
        :return: The refinement edge type.
        """
        pass

    @abstractmethod
    def get_alternative_edge(self, node: ast.AST) -> RDREdge:
        """
        :param node: The current AST node to determine the alternative edge type from.
        :return: The alternative edge type.
        """
        pass

    @abstractmethod
    def get_new_rule_type(self, node: ast.AST) -> Type[Rule]:
        """
        Get the new rule type to create.
        :param node: The current AST node to determine the rule type from.
        :return: The new rule type.
        """
        pass

    @abstractmethod
    def update_current_parent(self, new_node: Rule):
        """
        Update the current parent rule with the new node.
        :param new_node: The new node to set as the current parent.
        """
        pass

    def visit_Return(self, node):
        """Handles return statements as leaf nodes."""
        if isinstance(node.value, ast.Call):
            return_value = node.value.func.id
        else:
            return_value = ast.literal_eval(node.value)
        if self.current_parent is None:
            self.default_conclusion = return_value
        else:
            self.current_parent.conclusion = return_value

    def get_condition_name(self, node):
        """Extracts the condition function name from an AST expression."""
        if isinstance(node, ast.Call) and isinstance(node.func, ast.Name):
            return node.func.id
        return None

class SingleClassTreeBuilder(TreeBuilder):
    """Parses an AST of generated SingleClassRDR classifier and reconstructs the rdr tree."""

    def get_new_rule_type(self, node: ast.AST) -> Type[Rule]:
        return SingleClassRule

    def get_refinement_edge(self, node: ast.AST) -> RDREdge:
        return RDREdge.Refinement

    def get_alternative_edge(self, node: ast.AST) -> RDREdge:
        return RDREdge.Alternative

    def update_current_parent(self, new_node: Rule):
        if self.current_edge == RDREdge.Alternative:
            self.current_parent.alternative = new_node
        elif self.current_edge == RDREdge.Refinement:
            self.current_parent.refinement = new_node

    def process_else_statement(self, stmt: ast.AST):
        """Handles the else statement in the if-elif-else block."""
        if isinstance(stmt, ast.Return):
            self.current_parent = None
            self.visit_Return(stmt)
        else:
            raise ValueError(f"Unexpected statement in else block: {stmt}")


class MultiClassTreeBuilder(TreeBuilder):
    """Parses an AST of generated MultiClassRDR classifier and reconstructs the rdr tree."""

    def visit_If(self, stmt: ast.If):
        super().visit_If(stmt)
        if isinstance(self.current_parent, (MultiClassTopRule, MultiClassFilterRule)):
            self.current_parent.conclusion = self.current_parent.conditions.replace("conditions_", "conclusion_")

    def visit_Return(self, node):
        pass

    def get_new_rule_type(self, node: ast.AST) -> Type[Rule]:
        if self.current_edge == RDREdge.Refinement:
            return MultiClassStopRule
        elif self.current_edge == RDREdge.Filter:
            return MultiClassFilterRule
        elif self.current_edge in [RDREdge.Next, None]:
            return MultiClassTopRule
        elif self.current_edge == RDREdge.Alternative:
            return self.get_refinement_rule_type(node)
        else:
            raise ValueError(f"Unknown edge type: {self.current_edge}")

    def get_alternative_edge(self, node: ast.AST) -> RDREdge:
        if isinstance(self.current_parent, MultiClassTopRule):
            return RDREdge.Next
        else:
            return RDREdge.Alternative

    def get_refinement_edge(self, node: ast.AST) -> RDREdge:
        rule_type = self.get_refinement_rule_type(node)
        return self.get_refinement_edge_from_refinement_rule(rule_type)

    def get_refinement_edge_from_refinement_rule(self, rule_type: Type[Rule]) -> RDREdge:
        """
        :param rule_type: The type of the rule to determine the refinement edge from.
        :return: The refinement edge type based on the rule type.
        """
        if isinstance(self.current_parent, MultiClassRefinementRule):
            return RDREdge.Alternative
        if rule_type == MultiClassStopRule:
            return RDREdge.Refinement
        else:
            return RDREdge.Filter

    def get_refinement_rule_type(self, node: ast.AST) -> Type[Rule]:
        """
        :param node: The current AST node to determine the rule type from.
        :return: The rule type based on the node body.
        """
        for stmt in node.body:
            if len(node.body) == 1 and isinstance(stmt, ast.Pass):
                return MultiClassStopRule
            elif isinstance(stmt, ast.If):
                return self.get_refinement_rule_type(stmt)
            else:
                return MultiClassFilterRule
        raise ValueError(f"Could not determine the refinement rule type from the node: {node} as it has an empty body.")

    def update_current_parent(self, new_node: Rule):
        if isinstance(new_node, MultiClassRefinementRule):
            if isinstance(self.current_parent, MultiClassTopRule):
                new_node.top_rule = self.current_parent
            elif hasattr(self.current_parent, "top_rule"):
                new_node.top_rule = self.current_parent.top_rule
            else:
                raise ValueError(f"Could not set the top rule for the refinement rule: {new_node}")
        if self.current_edge in [RDREdge.Alternative, RDREdge.Next, None]:
            self.current_parent.alternative = new_node
        elif self.current_edge in [RDREdge.Refinement, RDREdge.Filter]:
            self.current_parent.refinement = new_node

    def process_else_statement(self, stmt: ast.AST):
        """Handles the else statement in the if-elif-else block."""
        pass

class RDRWithCodeWriter(RippleDownRules, ABC):

    @classmethod
    def read_rule_tree_from_python(cls, model_path: str, python_file_path: Optional[str] = None) -> Rule:
        """
        :param model_path: The path to the generated python file that contains the RDR classifier function.
        :param python_file_path: The path to the generated python file that contains the RDR classifier function.
        """
        if python_file_path is None:
            python_file_path = cls.get_generated_python_file_path(model_path)
        with open(python_file_path, "r") as f:
            source_code = f.read()

        tree = ast.parse(source_code)
        builder = cls.get_tree_builder_class()()

        # Find and process the function
        for node in tree.body:
            if isinstance(node, ast.FunctionDef) and node.name == "classify":
                builder.visit_FunctionDef(node)

        return builder.root

    @classmethod
    @abstractmethod
    def get_tree_builder_class(cls) -> Type[TreeBuilder]:
        """
        :return: The class that builds the rule tree from the generated python file.
        This should be either SingleClassTreeBuilder or MultiClassTreeBuilder.
        """
        pass

    @property
    def all_rules(self) -> List[Rule]:
        """
        Get all rules in the classifier.

        :return: A list of all rules in the classifier.
        """
        if self.start_rule is None:
            return []
        return [r for r in [self.start_rule] + list(self.start_rule.descendants) if r.conditions is not None]

    def update_from_python(self, model_dir: str, parent_package_name: Optional[str] = None,
                           python_file_path: Optional[str] = None, update_rule_tree: bool = False):
        """
        Update the rules from the generated python file, that might have been modified by the user.

        :param model_dir: The directory where the generated python file is located.
        :param parent_package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :param python_file_path: The path to the generated python file that contains the RDR classifier function.
        :param update_rule_tree: Whether to update the rule tree from the python file or not.
        """
        if update_rule_tree:
            self.start_rule = self.read_rule_tree_from_python(model_dir, python_file_path=python_file_path)
        all_rules = self.all_rules
        condition_func_names = [rule.generated_conditions_function_name for rule in all_rules]
        conclusion_func_names = [rule.generated_conclusion_function_name for rule in all_rules
                                 if not isinstance(rule, MultiClassStopRule)]
        all_func_names = condition_func_names + conclusion_func_names

        main_module, defs_module, cases_module = self.get_and_import_model_python_modules(
                                                        model_dir,
                                                        python_file_path=python_file_path,
                                                        parent_package_name=parent_package_name)
        self.generated_python_file_name = Path(main_module.__file__).name.replace(".py", "")

        self.update_rdr_metadata_from_python(main_module)

        functions_source = extract_function_source(defs_module.__file__,
                                                   all_func_names, include_signature=True)
        scope = extract_imports(defs_module.__file__, package_name=parent_package_name)

        cases_source, cases_scope = None, None
        if cases_module:
            with open(cases_module.__file__, "r") as f:
                cases_source = f.read()
            cases_scope = extract_imports(cases_module.__file__, package_name=parent_package_name)

        with open(main_module.__file__, "r") as f:
            main_source = f.read()
        main_scope = extract_imports(main_module.__file__, package_name=parent_package_name)
        attribute_name_line = [l for l in main_source.split('\n') if "attribute_name = " in l]
        conclusion_name = None
        if len(attribute_name_line) > 0:
            conclusion_name = eval(attribute_name_line[0].split('=')[-1].strip(), main_scope)

        for rule in all_rules:
            if rule.conditions is not None:
                conditions_wrapper_func_name = rule.generated_conditions_function_name
                user_input = functions_source[conditions_wrapper_func_name]
                user_input = '\n'.join(user_input.split("\n")[1:])  # Remove the function signature line
                rule.conditions = CallableExpression(user_input, (bool,), scope=scope)
                if cases_module:
                    try:
                        rule.corner_case_metadata = cases_module.__dict__[rule.generated_corner_case_object_name]
                    except KeyError:
                        case_def_lines = [l for l in cases_source.split('\n') if rule.generated_corner_case_object_name in l]
                        if len(case_def_lines) > 0:
                            case_def_line = case_def_lines[0].split('=')[-1].strip()
                            rule.corner_case_metadata = eval(case_def_line, cases_scope)

            if not isinstance(rule, MultiClassStopRule):
                if conclusion_name:
                    rule.conclusion_name = conclusion_name
                user_input = functions_source[rule.generated_conclusion_function_name]
                split_user_input = user_input.split("\n")
                user_input = '\n'.join(split_user_input[1:])
                conclusion_func = defs_module.__dict__.get(rule.generated_conclusion_function_name)
                if conclusion_func is None:
                    function_signature = split_user_input[0]
                    return_type_hint_str = function_signature.split('->')[-1].strip(' :')
                    return_type_hint = eval(return_type_hint_str, scope)
                    conclusion_type = get_type_from_type_hint(return_type_hint)
                else:
                    conclusion_type = get_function_return_type(conclusion_func)
                rule.conclusion = CallableExpression(user_input, conclusion_type, scope=scope,
                                                     mutually_exclusive=self.mutually_exclusive)

    @classmethod
    def get_and_import_model_python_modules(cls, model_dir: str,
                                            python_file_path: Optional[str] = None,
                                            parent_package_name: Optional[str] = None)\
            -> Tuple[ModuleType, ModuleType, ModuleType]:
        """
        Get and import the python modules that contain the RDR classifier function, definitions, and corner cases.

        :param model_dir: The path to the directory where the generated python files are located.
        :param python_file_path: The path to the generated python file that contains the RDR classifier function.
        :param parent_package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :return: A tuple containing the main module, defs module, and cases module.
        """
        if python_file_path is None:
            main_file_path = cls.get_generated_python_file_path(model_dir)
        else:
            main_file_path = python_file_path
        if not os.path.exists(main_file_path):
            raise ModuleNotFoundError(main_file_path)

        defs_file_path = main_file_path.replace(".py", "_defs.py")
        cases_path = main_file_path.replace(".py", "_cases.py")

        main_module, defs_module, cases_module = get_and_import_python_modules_in_a_package(
            [main_file_path, defs_file_path, cases_path], parent_package_name=parent_package_name)
        return main_module, defs_module, cases_module

    @abstractmethod
    def write_rules_as_source_code_to_file(self, rule: Rule, file, parent_indent: str = "",
                                           defs_file: Optional[str] = None, cases_file: Optional[str] = None,
                                           package_name: Optional[str] = None):
        """
        Write the rules as source code to a file.

        :param rule: The rule to write as source code.
        :param file: The file to write the source code to.
        :param parent_indent: The indentation of the parent rule.
        :param defs_file: The file to write the definitions to.
        :param cases_file: The file to write the cases to.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        pass

    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None):
        """
        Write the tree of rules as source code to a file.

        :param model_dir: The path to the directory to write the source code to.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        # Make sure the model directory exists and create an __init__.py file if it doesn't exist
        os.makedirs(model_dir, exist_ok=True)
        if not os.path.exists(model_dir + '/__init__.py'):
            with open(model_dir + '/__init__.py', 'w') as f:
                f.write("from . import *\n")

        # Set the file names for the generated python files
        file_name = model_dir + f"/{self.generated_python_file_name}.py"
        defs_file_name = model_dir + f"/{self.generated_python_defs_file_name}.py"
        cases_file_name = model_dir + f"/{self.generated_python_cases_file_name}.py"

        # Get the required imports for the main file and the defs file
        main_types, defs_types, corner_cases_types = self._get_types_to_import()
        imports = get_imports_from_types(main_types, file_name, package_name)
        defs_imports = get_imports_from_types(defs_types, defs_file_name, package_name)
        corner_cases_imports = get_imports_from_types(corner_cases_types, cases_file_name, package_name)

        defs_imports.append(f"from ripple_down_rules import *")
        # Add the imports to the defs file
        with open(defs_file_name, "w") as f:
            f.write('\n'.join(defs_imports) + "\n\n\n")

        # Add the imports to the cases file
        case_factory_import = get_imports_from_types([CaseFactoryMetaData], cases_file_name, package_name)
        corner_cases_imports.extend(case_factory_import)
        with open(cases_file_name, "w") as cases_f:
            cases_f.write("# This file contains the corner cases for the rules.\n")
            cases_f.write('\n'.join(corner_cases_imports) + "\n\n\n")

        # Add the imports, the attributes, and the function definition to the main file
        func_def = f"def classify(case: {self.case_type.__name__}, **kwargs) -> {self.conclusion_type_hint}:\n"
        with open(file_name, "w") as f:
            imports.append(f"from .{self.generated_python_defs_file_name} import *")
            f.write('\n'.join(imports) + "\n\n\n")
            f.write(f"attribute_name = '{self.attribute_name}'\n")
            f.write(f"conclusion_type = ({', '.join([ct.__name__ for ct in self.conclusion_type])},)\n")
            f.write(f"mutually_exclusive = {self.mutually_exclusive}\n")
            self.write_rdr_metadata_to_pyton_file(f)
            f.write(f"\n\n{func_def}")
            f.write(f"{' ' * 4}if not isinstance(case, Case):\n"
                    f"{' ' * 4}    case = create_case(case, max_recursion_idx=3)\n""")

        # Write the rules as source code to the main file
        self.write_rules_as_source_code_to_file(self.start_rule, file_name, " " * 4, defs_file=defs_file_name,
                                                cases_file=cases_file_name, package_name=package_name)

    @property
    @abstractmethod
    def conclusion_type_hint(self) -> str:
        """
        :return: The type hint of the conclusion of the rdr as a string.
        """
        pass

    def _get_types_to_import(self) -> Tuple[Set[Union[Type, Callable]], Set[Type], Set[Type]]:
        """
        :return: The types of the main, defs, and corner cases files of the RDR classifier that will be imported.
        """
        defs_types = set()
        cases_types = set()
        for rule in [self.start_rule] + list(self.start_rule.descendants):
            if not rule.conditions:
                continue
            for scope in [rule.conditions.scope, rule.conclusion.scope]:
                if scope is None:
                    continue
                defs_types.update(make_set(scope.values()))
            corner_case_types = rule.get_corner_case_types_to_import()
            if corner_case_types is not None:
                cases_types.update(corner_case_types)
        defs_types.add(self.case_type)
        main_types = set()
        main_types.add(self.case_type)
        main_types.update(make_set(self.conclusion_type))
        main_types.update({Union, Optional})
        defs_types.add(Union)
        main_types.update({Case, create_case})
        # main_types = main_types.difference(defs_types)
        return main_types, defs_types, cases_types

    @property
    def _default_generated_python_file_name(self) -> Optional[str]:
        """
        :return: The default generated python file name.
        """
        if self.start_rule is None or self.start_rule.conclusion is None:
            return None
        return f"{str_to_snake_case(self.case_name)}_{self.attribute_name}_{self.get_acronym().lower()}"

    @property
    def generated_python_defs_file_name(self) -> str:
        return f"{self.generated_python_file_name}_defs"

    @property
    def generated_python_cases_file_name(self) -> str:
        return f"{self.generated_python_file_name}_cases"

    @property
    def conclusion_type(self) -> Tuple[Type]:
        """
        :return: The type of the conclusion of the RDR classifier.
        """
        all_types = []
        for rule in self.all_rules:
            all_types.extend(list(rule.conclusion.conclusion_type))
        return tuple(set(all_types))

    @property
    def attribute_name(self) -> str:
        """
        :return: The name of the attribute that the classifier is classifying.
        """
        return self.start_rule.conclusion_name

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rule": self.start_rule.to_json(),
                "generated_python_file_name": self.generated_python_file_name,
                "name": self.name,
                "case_type": get_full_class_name(self.case_type) if self.case_type is not None else None,
                "case_name": self.case_name}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create an instance of the class from a json
        """
        start_rule = cls.start_rule_type().from_json(data["start_rule"])
        new_rdr = cls(start_rule=start_rule)
        if "generated_python_file_name" in data:
            new_rdr.generated_python_file_name = data["generated_python_file_name"]
        if "name" in data:
            new_rdr.name = data["name"]
        if "case_type" in data:
            new_rdr.case_type = get_type_from_string(data["case_type"])
        if "case_name" in data:
            new_rdr.case_name = data["case_name"]
        return new_rdr

    @staticmethod
    @abstractmethod
    def start_rule_type() -> Type[Rule]:
        """
        :return: The type of the starting rule of the RDR classifier.
        """
        pass


class SingleClassRDR(RDRWithCodeWriter):
    mutually_exclusive: bool = True
    """
    The output of the classification of this rdr negates all other possible outputs, there can only be one true value.
    """

    def __init__(self, default_conclusion: Optional[Any] = None, **kwargs):
        """
        :param start_rule: The starting rule for the classifier.
        :param default_conclusion: The default conclusion for the classifier if no rules fire.
        """
        super(SingleClassRDR, self).__init__(**kwargs)
        self.default_conclusion: Optional[Any] = default_conclusion

    @classmethod
    def get_tree_builder_class(cls) -> Type[TreeBuilder]:
        return SingleClassTreeBuilder

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CaseAttribute, CallableExpression, None]:
        """
        Classify a case, and ask the user for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category if provided.

        :param case_query: The case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        if case_query.default_value is not None and self.default_conclusion != case_query.default_value:
            self.default_conclusion = case_query.default_value

        pred = self.evaluate(case_query.case)
        if (not pred.fired and self.default_conclusion is None) or pred.conclusion(case_query.case) != case_query.target_value:
            expert.ask_for_conditions(case_query, pred)
            pred.fit_rule(case_query)

        return self.classify(case_query.case)

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if not self.start_rule:
            expert.ask_for_conditions(case_query)
            self.start_rule = SingleClassRule.from_case_query(case_query)

    def _classify(self, case: Case, modify_case: bool = False,
                  case_query: Optional[CaseQuery] = None) -> Optional[Any]:
        """
        Classify a case by recursively evaluating the rules until a rule fires or the last rule is reached.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case attributes with the conclusion or not.
        :param case_query: The case query containing the case and the target category to compare the case with.
        """
        pred = self.evaluate(case)
        conclusion = pred.conclusion(case) if pred is not None and pred.fired else self.default_conclusion
        if pred is not None and pred.fired:
            pred.contributed = True
            pred.last_conclusion = conclusion
            if case_query is not None:
                pred.contributed_to_case_query = True
        if pred is not None and pred.fired and case_query is not None:
            if pred.corner_case_metadata is None and conclusion is not None \
                    and type(conclusion) in case_query.core_attribute_type:
                pred.corner_case_metadata = CaseFactoryMetaData.from_case_query(case_query)
        return conclusion

    def evaluate(self, case: Case) -> SingleClassRule:
        """
        Evaluate the starting rule on a case.
        """
        matched_rule = self.start_rule(case) if self.start_rule is not None else None
        return matched_rule if matched_rule is not None else self.start_rule

    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None):
        super()._write_to_python(model_dir, package_name=package_name)
        with open(model_dir + f"/{self.generated_python_file_name}.py", "a") as f:
            f.write(f"{' ' * 4}else:\n{' ' * 8}return {self.default_conclusion}\n")

    def write_rules_as_source_code_to_file(self, rule: SingleClassRule, filename: str, parent_indent: str = "",
                                           defs_file: Optional[str] = None, cases_file: Optional[str] = None,
                                           package_name: Optional[str] = None):
        """
        Write the rules as source code to a file.
        """
        if rule.conditions:
            rule.write_corner_case_as_source_code(cases_file, package_name=package_name)
            if_clause = rule.write_condition_as_source_code(parent_indent, defs_file)
            with open(filename, "a") as file:
                file.write(if_clause)
            if rule.refinement:
                self.write_rules_as_source_code_to_file(rule.refinement, filename, parent_indent + "    ",
                                                        defs_file=defs_file, cases_file=cases_file,
                                                        package_name=package_name)

            conclusion_call = rule.write_conclusion_as_source_code(parent_indent, defs_file)
            with open(filename, "a") as file:
                file.write(conclusion_call)

            if rule.alternative:
                self.write_rules_as_source_code_to_file(rule.alternative, filename, parent_indent, defs_file=defs_file,
                                                        cases_file=cases_file, package_name=package_name)

    @property
    def conclusion_type_hint(self) -> str:
        all_types = set(list(self.conclusion_type) + [type(self.default_conclusion)])
        if NoneType in all_types:
            return f"Optional[{', '.join([t.__name__ for t in all_types if t is not NoneType])}]"
        return f"Union[{', '.join([t.__name__ for t in all_types])}]"

    def _get_types_to_import(self) -> Tuple[Set[Type], Set[Type], Set[Type]]:
        main_types, def_types, case_types = super()._get_types_to_import()
        main_types.add(type(self.default_conclusion))
        def_types.add(type(self.default_conclusion))
        if self.default_conclusion is None:
            main_types.add(Optional)
            def_types.add(Optional)
        return main_types, def_types, case_types

    @property
    def conclusion_type(self) -> Tuple[Type]:
        if self.default_conclusion is not None:
            return (type(self.default_conclusion),)
        return super().conclusion_type

    @staticmethod
    def start_rule_type() -> Type[Rule]:
        """
        :return: The type of the starting rule of the RDR classifier.
        """
        return SingleClassRule


class MultiClassRDR(RDRWithCodeWriter):
    """
    A multi class ripple down rules classifier, which can draw multiple conclusions for a case.
    This is done by going through all rules and checking if they fire or not, and adding stopping rules if needed,
    when wrong conclusions are made to stop these rules from firing again for similar cases.
    """
    evaluated_rules: Optional[List[Rule]] = None
    """
    The evaluated rules in the classifier for one case.
    """
    conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the case belongs to.
    """
    stop_rule_conditions: Optional[CallableExpression] = None
    """
    The conditions of the stopping rule if needed.
    """
    mutually_exclusive: bool = False
    """
    The output of the classification of this rdr allows for more than one true value as conclusion.
    """

    def __init__(self, start_rule: Optional[MultiClassTopRule] = None,
                 mode: MCRDRMode = MCRDRMode.StopOnly, **kwargs):
        """
        :param start_rule: The starting rules for the classifier.
        :param mode: The mode of the classifier, either StopOnly or StopPlusRule, or StopPlusRuleCombined.
        """
        super(MultiClassRDR, self).__init__(start_rule, **kwargs)
        self.mode: MCRDRMode = mode

    @classmethod
    def get_tree_builder_class(cls) -> Type[TreeBuilder]:
        return MultiClassTreeBuilder

    def _classify(self, case: Union[Case, SQLTable], modify_case: bool = False,
                  case_query: Optional[CaseQuery] = None) -> Set[Any]:
        evaluated_rule = self.start_rule
        self.conclusions = []
        while evaluated_rule:
            next_rule = evaluated_rule(case)
            if evaluated_rule.fired:
                rule_conclusion = evaluated_rule.conclusion(case)
                if evaluated_rule.corner_case_metadata is None and case_query is not None:
                    if rule_conclusion is not None and len(make_list(rule_conclusion)) > 0 \
                            and any(
                        ct in case_query.core_attribute_type for ct in map(type, make_list(rule_conclusion))):
                        evaluated_rule.corner_case_metadata = CaseFactoryMetaData.from_case_query(case_query)
                if rule_conclusion is not None and any(make_list(rule_conclusion)):
                    evaluated_rule.contributed = True
                    evaluated_rule.last_conclusion = rule_conclusion
                    if case_query is not None:
                        rule_conclusion_types = set(map(type, make_list(rule_conclusion)))
                        if are_results_subclass_of_types(rule_conclusion_types, case_query.core_attribute_type):
                            evaluated_rule.contributed_to_case_query = True
                self.add_conclusion(rule_conclusion)
            evaluated_rule = next_rule
        return make_set(self.conclusions)

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None
                  , **kwargs) -> Set[Union[CaseAttribute, CallableExpression, None]]:
        """
        Classify a case, and ask the user for stopping rules or classifying rules if the classification is incorrect
         or missing by comparing the case with the target category if provided.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions or for extra conclusions.
        :return: The conclusions that the case belongs to.
        """
        self.conclusions = []
        self.stop_rule_conditions = None
        evaluated_rule = self.start_rule
        target_value = make_set(case_query.target_value)
        while evaluated_rule:
            next_rule = evaluated_rule(case_query.case)
            rule_conclusion = evaluated_rule.conclusion(case_query.case)

            if evaluated_rule.fired:
                if not make_set(rule_conclusion).issubset(target_value):
                    # Rule fired and conclusion is different from target
                    self.stop_wrong_conclusion_else_add_it(case_query, expert, evaluated_rule)
                else:
                    # Rule fired and target is correct or there is no target to compare
                    self.add_conclusion(rule_conclusion)

            if not next_rule:
                if not make_set(target_value).issubset(make_set(self.conclusions)):
                    # Nothing fired and there is a target that should have been in the conclusions
                    self.add_rule_for_case(case_query, expert)
                    # Have to check all rules again to make sure only this new rule fires
                    next_rule = self.start_rule
            evaluated_rule = next_rule
        return self.conclusions

    def write_rules_as_source_code_to_file(self, rule: Union[MultiClassTopRule, MultiClassStopRule],
                                           filename: str, parent_indent: str = "", defs_file: Optional[str] = None,
                                           cases_file: Optional[str] = None, package_name: Optional[str] = None):
        if rule == self.start_rule:
            with open(filename, "a") as file:
                file.write(f"{parent_indent}conclusions = set()\n")
        if rule.conditions:
            rule.write_corner_case_as_source_code(cases_file, package_name=package_name)
            if_clause = rule.write_condition_as_source_code(parent_indent, defs_file)
            with open(filename, "a") as file:
                file.write(if_clause)
            conclusion_indent = parent_indent
            if hasattr(rule, "refinement") and rule.refinement:
                self.write_rules_as_source_code_to_file(rule.refinement, filename, parent_indent + "    ",
                                                        defs_file=defs_file, cases_file=cases_file,
                                                        package_name=package_name)
                conclusion_indent = parent_indent + " " * 4
                with open(filename, "a") as file:
                    file.write(f"{conclusion_indent}else:\n")

            conclusion_call = rule.write_conclusion_as_source_code(conclusion_indent, defs_file)
            with open(filename, "a") as file:
                file.write(conclusion_call)

            if rule.alternative:
                self.write_rules_as_source_code_to_file(rule.alternative, filename, parent_indent, defs_file=defs_file,
                                                        cases_file=cases_file, package_name=package_name)
            elif isinstance(rule, MultiClassTopRule):
                with open(filename, "a") as file:
                    file.write(f"{parent_indent}return conclusions\n")

    @property
    def conclusion_type_hint(self) -> str:
        conclusion_types = [ct.__name__ for ct in self.conclusion_type if ct not in [list, set]]
        if len(conclusion_types) == 1:
            return f"Set[{conclusion_types[0]}]"
        else:
            return f"Set[Union[{', '.join(conclusion_types)}]]"

    def _get_types_to_import(self) -> Tuple[Set[Union[Type, Callable]], Set[Type], Set[Type]]:
        main_types, defs_types, cases_types = super()._get_types_to_import()
        main_types.add(get_an_updated_case_copy)
        main_types.update({Set, make_set})
        defs_types.update({List, Set})
        return main_types, defs_types, cases_types

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if not self.start_rule:
            conditions = expert.ask_for_conditions(case_query)
            self.start_rule: MultiClassTopRule = MultiClassTopRule.from_case_query(case_query)

    @property
    def last_top_rule(self) -> Optional[MultiClassTopRule]:
        """
        Get the last top rule in the tree.
        """
        if not self.start_rule.furthest_alternative:
            return self.start_rule
        else:
            return self.start_rule.furthest_alternative[-1]

    def stop_wrong_conclusion_else_add_it(self, case_query: CaseQuery, expert: Expert,
                                          evaluated_rule: MultiClassTopRule):
        """
        Stop a wrong conclusion by adding a stopping rule.
        """
        rule_conclusion = evaluated_rule.conclusion(case_query.case)
        stop: bool = False
        add_filter_rule: bool = False
        if is_value_conflicting(rule_conclusion, case_query.target_value):
            if make_set(case_query.target_value).issubset(rule_conclusion):
                add_filter_rule = True
            else:
                stop = True
        elif make_set(case_query.core_attribute_type).issubset(make_set(evaluated_rule.conclusion.conclusion_type)):
            if make_set(case_query.target_value).issubset(rule_conclusion):
                add_filter_rule = True

        if not stop:
            self.add_conclusion(rule_conclusion)
        if stop or add_filter_rule:
            refinement_type = MultiClassStopRule if stop else MultiClassFilterRule
            self.stop_or_filter_conclusion(case_query, expert, evaluated_rule, refinement_type=refinement_type)

    def stop_or_filter_conclusion(self, case_query: CaseQuery,
                                  expert: Expert, evaluated_rule: MultiClassTopRule,
                                  refinement_type: Type[MultiClassRefinementRule] = MultiClassStopRule):
        """
        Stop a conclusion by adding a stopping rule.

        :param case_query: The case query to stop the conclusion for.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param evaluated_rule: The evaluated rule to ask the expert about.
        :param refinement_type: The refinement type to use.
        """
        conditions = expert.ask_for_conditions(case_query, evaluated_rule)
        evaluated_rule.fit_rule(case_query, refinement_type=refinement_type)
        if refinement_type is MultiClassStopRule:
            if self.mode == MCRDRMode.StopPlusRule:
                self.stop_rule_conditions = conditions
            if self.mode == MCRDRMode.StopPlusRuleCombined:
                new_top_rule_conditions = conditions.combine_with(evaluated_rule.conditions)
                case_query.conditions = new_top_rule_conditions
                self.add_top_rule(case_query)

    def add_rule_for_case(self, case_query: CaseQuery, expert: Expert):
        """
        Add a rule for a case that has not been classified with any conclusion.

        :param case_query: The case query to add the rule for.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if self.stop_rule_conditions and self.mode == MCRDRMode.StopPlusRule:
            conditions = self.stop_rule_conditions
            self.stop_rule_conditions = None
            case_query.conditions = conditions
        else:
            conditions = expert.ask_for_conditions(case_query)
        self.add_top_rule(case_query)

    def add_conclusion(self, rule_conclusion: List[Any]) -> None:
        """
        Add the conclusion of the evaluated rule to the list of conclusions.

        :param rule_conclusion: The conclusion of the evaluated rule, which can be a single conclusion
         or a set of conclusions.
        """
        conclusion_types = [type(c) for c in self.conclusions]
        if type(rule_conclusion) not in conclusion_types:
            self.conclusions.extend(make_list(rule_conclusion))
        else:
            same_type_conclusions = [c for c in self.conclusions if type(c) == type(rule_conclusion)]
            combined_conclusion = rule_conclusion if isinstance(rule_conclusion, set) \
                else {rule_conclusion}
            combined_conclusion = copy(combined_conclusion)
            for c in same_type_conclusions:
                combined_conclusion.update(c if isinstance(c, set) else make_set(c))
                self.conclusions.remove(c)
            self.conclusions.extend(make_list(combined_conclusion))

    def add_top_rule(self, case_query: CaseQuery):
        """
        Add a top rule to the classifier, which is a rule that is always checked and is part of the start_rules list.

        :param case_query: The case query to add the top rule for.
        """
        self.start_rule.alternative = MultiClassTopRule.from_case_query(case_query)

    @staticmethod
    def start_rule_type() -> Type[Rule]:
        """
        :return: The type of the starting rule of the RDR classifier.
        """
        return MultiClassTopRule


class GeneralRDR(RippleDownRules):
    """
    A general ripple down rules classifier, which can draw multiple conclusions for a case, but each conclusion is part
    of a set of mutually exclusive conclusions. Whenever a conclusion is made, the classification restarts from the
    starting rule, and all the rules that belong to the class of the made conclusion are not checked again. This
    continues until no more rules can be fired. In addition, previous conclusions can be used as conditions or input to
    the next classification/cycle.
    Another possible mode is to have rules that are considered final, when fired, inference will not be restarted,
     and only a refinement can be made to the final rule, those can also be used in another SCRDR of their own that
     gets called when the final rule fires.
    """

    def __init__(self, category_rdr_map: Optional[Dict[str, Union[SingleClassRDR, MultiClassRDR]]] = None,
                 **kwargs):
        """
        :param category_rdr_map: A map of case attribute names to ripple down rules classifiers,
        where each category is a parent category that has a set of mutually exclusive (in case of SCRDR) child
        categories, e.g. {'species': SCRDR, 'habitats': MCRDR}, where 'species' and 'habitats' are attribute names
        for a case of type Animal, while SCRDR and MCRDR are SingleClass and MultiClass ripple down rules classifiers.
        Species can have values like Mammal, Bird, Fish, etc. which are mutually exclusive, while Habitat can have
        values like Land, Water, Air, etc., which are not mutually exclusive due to some animals living more than one
        habitat.
        """
        self.start_rules_dict: Dict[str, Union[SingleClassRDR, MultiClassRDR]] \
            = category_rdr_map if category_rdr_map else {}
        super(GeneralRDR, self).__init__(**kwargs)
        self.all_figs: List[Figure] = [sr.fig for sr in self.start_rules_dict.values()]

    @classmethod
    def from_python(cls, model_dir: str, python_file_path: Optional[str] = None,
                    parent_package_name: Optional[str] = None) -> Self:
        """
        Create an instance of the class from a python file.

        :param model_dir: The path to the directory containing the python file.
        :param python_file_path: The path to the python file, if not provided, it will be generated from the model_dir.
        :param parent_package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :return: An instance of the class.
        """
        grdr = cls()
        grdr.update_from_python(model_dir, parent_package_name=parent_package_name, python_file_path=python_file_path,
                                update_rule_tree=True)
        return grdr

    @classmethod
    def get_rdr_type_from_acronym(cls, acronym: str) -> Type[Union[SingleClassRDR, MultiClassRDR]]:
        """
        Get the type of the ripple down rules classifier from the acronym.

        :param acronym: The acronym of the ripple down rules classifier.
        :return: The type of the ripple down rules classifier.
        """
        acronym = acronym.lower()
        if acronym == "scrdr":
            return SingleClassRDR
        elif acronym == "mcrdr":
            return MultiClassRDR
        else:
            raise ValueError(f"Unknown RDR type acronym: {acronym}")

    def add_rdr(self, rdr: Union[SingleClassRDR, MultiClassRDR], case_query: Optional[CaseQuery] = None):
        """
        Add a ripple down rules classifier to the map of classifiers.

        :param rdr: The ripple down rules classifier to add.
        :param case_query: The case query to add the classifier for.
        """
        name = case_query.attribute_name if case_query else rdr.name
        self.start_rules_dict[name] = rdr

    @property
    def start_rule(self) -> Optional[Union[SingleClassRule, MultiClassTopRule]]:
        return self.start_rules[0] if self.start_rules_dict else None

    @start_rule.setter
    def start_rule(self, value: Union[SingleClassRDR, MultiClassRDR]):
        if value:
            self.start_rules_dict[value.attribute_name] = value

    @property
    def start_rules(self) -> List[Union[SingleClassRule, MultiClassTopRule]]:
        return [rdr.start_rule for rdr in self.start_rules_dict.values()]

    def _classify(self, case: Any, modify_case: bool = False,
                  case_query: Optional[CaseQuery] = None) -> Optional[Dict[str, Any]]:
        """
        Classify a case by going through all RDRs and adding the categories that are classified, and then restarting
        the classification until no more categories can be added.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case or create a copy and modify it.
        :param case_query: The case query containing the case and the target category to compare the case with.
        :return: The categories that the case belongs to.
        """
        return general_rdr_classify(self.start_rules_dict, case, modify_original_case=modify_case,
                                    case_query=case_query)

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Dict[str, Any]:
        """
        Fit the GRDR on a case, if the target is a new type of category, a new RDR is created for it,
        else the existing RDR of that type will be fitted on the case, and then classification is done and all
        concluded categories are returned. If the category is mutually exclusive, an SCRDR is created, else an MCRDR.
        In case of SCRDR, multiple conclusions of the same type replace each other, in case of MCRDR, they are added if
        they are accepted by the expert, and the attribute of that category is represented in the case as a set of
        values.

        :param case_query: The query containing the case to classify and the target category to compare the case
        with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The categories that the case belongs to.
        """
        case_query_cp = copy(case_query)
        self.classify(case_query_cp.case, modify_case=True)
        case_query_cp.update_target_value()

        self.start_rules_dict[case_query_cp.attribute_name].fit_case(case_query_cp, expert, **kwargs)

        return self.classify(case_query.case)

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if case_query.attribute_name not in self.start_rules_dict:
            new_rdr = self.initialize_new_rdr_for_attribute(case_query)
            self.add_rdr(new_rdr, case_query)

    @staticmethod
    def initialize_new_rdr_for_attribute(case_query: CaseQuery):
        """
        Initialize the appropriate RDR type for the target.
        """
        return SingleClassRDR(default_conclusion=case_query.default_value) if case_query.mutually_exclusive \
            else MultiClassRDR()

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rules": {name: rdr.to_json() for name, rdr in self.start_rules_dict.items()}
            , "generated_python_file_name": self.generated_python_file_name,
                "name": self.name,
                "case_type": get_full_class_name(self.case_type) if self.case_type is not None else None,
                "case_name": self.case_name}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> GeneralRDR:
        """
        Create an instance of the class from a json
        """
        start_rules_dict = {}
        for k, v in data["start_rules"].items():
            start_rules_dict[k] = get_type_from_string(v['_type']).from_json(v)
        new_rdr = cls(category_rdr_map=start_rules_dict)
        if "generated_python_file_name" in data:
            new_rdr.generated_python_file_name = data["generated_python_file_name"]
        if "name" in data:
            new_rdr.name = data["name"]
        if "case_type" in data:
            new_rdr.case_type = get_type_from_string(data["case_type"])
        if "case_name" in data:
            new_rdr.case_name = data["case_name"]
        return new_rdr

    def update_from_python(self, model_dir: str, parent_package_name: Optional[str] = None,
                           python_file_path: Optional[str] = None,
                           update_rule_tree: bool = False) -> None:
        """
        Update the rules from the generated python file, that might have been modified by the user.

        :param model_dir: The directory where the model is stored.
        :param parent_package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :param python_file_path: The path to the python file, if not provided, it will be generated from the model_dir.
        :param update_rule_tree: Whether to update the rule tree from the python file or not.
        """
        if update_rule_tree:
            if python_file_path is None:
                main_python_file_path = self.get_generated_python_file_path(model_dir)
            else:
                main_python_file_path = python_file_path
            main_module = get_and_import_python_module(main_python_file_path, parent_package_name=parent_package_name)
            classifiers_dict = main_module.classifiers_dict
            self.start_rules_dict = {}
            for rdr_name, rdr_module in classifiers_dict.items():
                rdr_acronym = rdr_module.__name__.split('_')[-1]
                rdr_type = self.get_rdr_type_from_acronym(rdr_acronym)
                rdr_model_path = main_python_file_path.replace('_rdr.py', f'_{rdr_name}_{rdr_acronym}.py')
                rdr = rdr_type.from_python(model_dir, python_file_path=rdr_model_path,
                                           parent_package_name=parent_package_name)
                self.start_rules_dict[rdr_name] = rdr

            self.update_rdr_metadata_from_python(main_module)
        else:
            for rdr in self.start_rules_dict.values():
                rdr.update_from_python(model_dir, parent_package_name=parent_package_name)

    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None) -> None:
        """
        Write the tree of rules as source code to a file.

        :param model_dir: The directory where the model is stored.
        :param package_name: The name of the package that contains the RDR classifier function.
        """
        for rdr in self.start_rules_dict.values():
            rdr._write_to_python(model_dir, package_name=package_name)
        func_def = f"def classify(case: {self.case_type.__name__}, **kwargs) -> {self.conclusion_type_hint}:\n"
        file_path = model_dir + f"/{self.generated_python_file_name}.py"
        with open(file_path, "w") as f:
            f.write(self._get_imports(file_path=file_path, package_name=package_name) + "\n\n")
            self.write_rdr_metadata_to_pyton_file(f)
            f.write("classifiers_dict = dict()\n")
            for rdr_key, rdr in self.start_rules_dict.items():
                f.write(f"classifiers_dict['{rdr_key}'] = {self.rdr_key_to_function_name(rdr_key)}\n")
            f.write("\n\n")
            f.write(func_def)
            f.write(f"{' ' * 4}if not isinstance(case, Case):\n"
                    f"{' ' * 4}    case = create_case(case, max_recursion_idx=3)\n""")
            f.write(f"{' ' * 4}return general_rdr_classify(classifiers_dict, case, **kwargs)\n")

    @property
    def _default_generated_python_file_name(self) -> Optional[str]:
        """
        :return: The default generated python file name.
        """
        if self.start_rule is None or self.start_rule.conclusion is None:
            return None
        return f"{str_to_snake_case(self.case_name)}_rdr".lower()

    @property
    def conclusion_type_hint(self) -> str:
        return f"Dict[str, Any]"

    def _get_imports(self, file_path: Optional[str] = None, package_name: Optional[str] = None) -> str:
        """
        Get the imports needed for the generated python file.

        :param file_path: The path to the file where the imports will be written, if None, the imports will be absolute.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :return: The imports needed for the generated python file.
        """
        all_types = set()
        # add type hints
        all_types.update({Dict, Any})
        # import rdr type
        all_types.add(general_rdr_classify)
        # add case type
        all_types.update({Case, create_case, self.case_type})
        # get the imports from the types
        imports = get_imports_from_types(all_types, target_file_path=file_path, package_name=package_name)
        # add rdr python generated functions.
        for rdr_key, rdr in self.start_rules_dict.items():
            imports.append(
                f"from . import {rdr.generated_python_file_name} as {self.rdr_key_to_function_name(rdr_key)}")
        return '\n'.join(imports)

    @staticmethod
    def rdr_key_to_function_name(rdr_key: str) -> str:
        """
        Convert the RDR key to a function name.

        :param rdr_key: The RDR key to convert.
        :return: The function name.
        """
        return rdr_key.replace(".", "_").lower() + "_classifier"



from dataclasses import dataclass

from typing_extensions import List

from relational_model import Part, Robot as OGRobot, PhysicalObject
from ripple_down_rules import RDRDecorator


@dataclass(unsafe_hash=True)
class Robot(OGRobot):
    containment_rdr: RDRDecorator = RDRDecorator("./", (PhysicalObject,), False,
                                                 fit=False)

    @containment_rdr.decorator
    def get_contained_objects(self) -> List[PhysicalObject]:
        """
        Returns the contained objects of the robot.
        """
        ...

def my_robot_factory() -> Robot:
    """
    Factory function to create a simple robot with parts and containment relationships.
    """
    part_a = Part(name="A")
    part_b = Part(name="B")
    part_c = Part(name="C")
    robot = Robot("pr2", parts=[part_a])
    part_a.contained_objects = [part_b]
    part_b.contained_objects = [part_c]

    return robot


from typing_extensions import Any, Dict
from ripple_down_rules.helpers import general_rdr_classify
from ripple_down_rules.datastructures.case import Case, create_case
from . import robot_get_contained_objects_output__mcrdr as output__classifier

name = 'output_'
case_type = Dict
case_name = 'Robot_get_contained_objects'
classifiers_dict = dict()
classifiers_dict['output_'] = output__classifier


def classify(case: Dict, **kwargs) -> Dict[str, Any]:
    if not isinstance(case, Case):
        case = create_case(case, max_recursion_idx=3)
    return general_rdr_classify(classifiers_dict, case, **kwargs)



__version__ = "0.6.70"

import logging
import sys

logger = logging.Logger("rdr")
logger.setLevel(logging.INFO)

try:
    from PyQt6.QtWidgets import QApplication
    app = QApplication(sys.argv)
except ImportError:
    pass


# Trigger patch
try:
    from .predicates import *
    from .datastructures.tracked_object import TrackedObjectMixin
    from .datastructures.dataclasses import CaseQuery
    from .rdr_decorators import RDRDecorator
    from .rdr import MultiClassRDR, SingleClassRDR, GeneralRDR
    import ripple_down_rules_meta._apply_overrides
except ImportError:
    pass


from decorator_model import Robot
from relational_model import Part, PhysicalObject
from ripple_down_rules.helpers import enable_gui


# Define a simple robot with parts and containment relationships
part_a = Part(name="A")
part_b = Part(name="B")
part_c = Part(name="C")
robot = Robot("pr2", parts=[part_a])
part_a.contained_objects = [part_b]
part_b.contained_objects = [part_c]

# Optional: Use the GUI.r
enable_gui()

# Create a GeneralRDR instance and fit it to the case query
robot.containment_rdr.fit = True
robot.get_contained_objects()

# Classify the robot to check if it contains part_b
robot.containment_rdr.fit = False
result = robot.get_contained_objects()
assert result == [part_b]



from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    return case.species=='mammal' and case.aquatic==0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    return case.species=='fish'



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    return case.species=='bird' and case.legs>0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    return case.species=='molusc' and case.aquatic==0



'===New Answer==='


from typing_extensions import Any, Callable, List, Optional, Tuple, Type
from test.datasets import Habitat, Species, load_zoo_cases
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import GeneralRDR, MultiClassRDR, SingleClassRDR
from ripple_down_rules.utils import make_set
from test.test_helpers.helpers import get_fit_grdr, get_fit_mcrdr, get_fit_scrdr, get_habitat
from pandas.core.frame import DataFrame

def conditions_for_animal_habitats_of_type_habitat(case: DataFrame) -> bool:
    """Get conditions on whether it's possible to conclude a value for Animal.habitats  of type Habitat."""
    # Write your code here
    return case.species=='molusc' and case.aquatic==1



'===New Answer==='





import os
from os.path import dirname

from IPython import embed
from hypothesis.strategies import recursive

from ripple_down_rules import *
from ripple_down_rules.datastructures.tracked_object import X
from .datasets import Drawer, Handle, Cabinet, View, WorldEntity, Body, Connection, Container


def test_construct_class_hierarchy():
    TrackedObjectMixin._reset_dependency_graph()
    TrackedObjectMixin.make_class_dependency_graph(composition=False)
    Drawer.to_dot(os.path.join(dirname(__file__), "dependency_graph"))
    assert len(Drawer._dependency_graph.nodes()) == 20
    assert len(Drawer._dependency_graph.edges()) == 17


def test_construct_class_composition():
    TrackedObjectMixin._reset_dependency_graph()
    TrackedObjectMixin.make_class_dependency_graph(composition=True)
    Drawer.to_dot(os.path.join(dirname(__file__), "dependency_graph"))
    assert len(Drawer._dependency_graph.nodes()) == 20
    assert len(Drawer._dependency_graph.edges()) == 22
    Drawer.to_dot(os.path.join(dirname(__file__), "dependency_graph"))


# @pytest.mark.skip("Not Implemented yet")
def test_construct_class_composition_and_dependency():
    TrackedObjectMixin._reset_dependency_graph()
    TrackedObjectMixin.make_class_dependency_graph(composition=True)
    assert next(has(Drawer, Handle))
    assert next(has(Cabinet, Drawer))
    assert list(has(Cabinet, X)) == [(Cabinet, Drawer), (Cabinet, Container)]
    assert list(has(Cabinet, X, recursive=True)) == [(Cabinet, Drawer), (Cabinet, Container), (Cabinet, Container), (Cabinet, Handle)]
    assert list(has(X, Handle)) == [(Drawer, Handle)]
    assert list(has(X, Handle, recursive=True)) == [(Drawer, Handle), (Cabinet, Handle)]
    assert isA(Cabinet, View)
    assert isA(Cabinet, WorldEntity)
    assert not list(has(Cabinet, Handle))
    assert next(has(Cabinet, Handle, recursive=True))
    assert next(has(Cabinet, Body))
    assert next(has(Cabinet, WorldEntity))
    assert not list(has(Cabinet, Connection, recursive=True))



from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from typing import Generator, ClassVar

from typing_extensions import Type, TYPE_CHECKING, Tuple, Dict, List

from .datastructures.enums import InferMode
from .datastructures.tracked_object import TrackedObjectMixin, Direction, Relation

if TYPE_CHECKING:
    pass


@dataclass(eq=False)
class Predicate(TrackedObjectMixin, ABC):

    cached_results: ClassVar[Dict[Tuple, List]] = defaultdict(list)
    """
    Cached results between retrievals, makes future retrievals faster,
    """
    inferred_once: ClassVar[bool] = None
    """
    If inference has been performed at least once.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.inferred_once = False

    @classmethod
    @abstractmethod
    def relation(cls) -> Relation:
        """
        The relation type of the predicate.
        """

    def __call__(self, *args, infer: InferMode = InferMode.Auto, **kwargs):
        if self.should_infer(mode=infer):
            self.infer(*args, **kwargs)
            self.cached_results.clear()
            self.__class__.inferred_once = True
        key = (args, tuple(sorted(kwargs.items())))
        if key in self.cached_results:
            yield from self.cached_results[key]
        else:
            for value in self.retrieve(*args, **kwargs):
                self.cached_results[key].append(value)
                yield value

    @classmethod
    def should_infer(cls, mode: InferMode) -> bool:
        """
        Determine if the predicate relations should be inferred or just retrieve current relations.

        :param mode: The infer mode of the predicate (could be Auto, Always, Never)
        """
        if mode == InferMode.Always:
            return True
        elif mode == InferMode.Never:
            return False
        else:
            return cls._should_infer()

    @classmethod
    @abstractmethod
    def _should_infer(cls) -> bool:
        """
        Predicate specific reasoning on when to infer relations.
        """

    @classmethod
    @abstractmethod
    def infer(cls, *args, **kwargs):
        """
        Evaluate the predicate with the given arguments.
        This method should be implemented by subclasses.
        """

    @classmethod
    def retrieve(cls, node1_type: Type[TrackedObjectMixin],
                 node2_type: Type[TrackedObjectMixin], recursive: bool = False, is_reversed: bool = False) \
            -> Generator[Tuple[Type[TrackedObjectMixin], Type[TrackedObjectMixin]], None, None]:
        if node1_type is not TrackedObjectMixin:
            direction = Direction.INBOUND.value if is_reversed else Direction.OUTBOUND.value
            owner_idx = node1_type._my_graph_idx()
            neighbors = cls._dependency_graph.adj_direction(owner_idx, direction)
            neighbor_generator = ((owner_idx, n) for n, e in neighbors.items()
                                  if (e == cls.relation() and issubclass(cls._dependency_graph.get_node_data(n),
                                                                         node2_type))
                                  or (e == Relation.isA and any(cls.retrieve(cls._dependency_graph.get_node_data(n),
                                                                             node2_type)))
                                  )
            latest_results = []
            for v in neighbor_generator:
                res = (cls._dependency_graph.get_node_data(v[0]), cls._dependency_graph.get_node_data(v[1]))
                latest_results.append(v)
                yield res
            if recursive:
                for n in [n for n, e in neighbors.items() if e == cls.relation()]:
                    for v in cls.retrieve(cls._dependency_graph.get_node_data(n), node2_type, recursive=True,
                                          is_reversed=is_reversed):
                        yield node1_type, v[1]
        elif node2_type is not TrackedObjectMixin:
            # owner type is TrackedObjectMixin
            yield from map(lambda t: (t[1], t[0]),
                           cls.retrieve(node2_type, node1_type, recursive=recursive, is_reversed=True))
        else:
            # both are TrackedObjectMixin
            yield from map(
                lambda t: (cls._dependency_graph.get_node_data(t[0]), cls._dependency_graph.get_node_data(t[1])),
                cls._edges[cls.relation()])

    def __hash__(self):
        return hash(self.__class__.__name__)

    def __eq__(self, other):
        if not isinstance(other, Predicate):
            return False
        return self.__class__ == other.__class__


@dataclass(eq=False)
class IsA(Predicate):
    """
    A predicate that checks if an object type is a subclass of another object type.
    """

    @classmethod
    def relation(cls):
        return Relation.isA

    @classmethod
    def infer(cls, child_type: Type[TrackedObjectMixin], parent_type: Type[TrackedObjectMixin]):
        TrackedObjectMixin.make_class_dependency_graph(composition=False)

    @classmethod
    def _should_infer(cls) -> bool:
        return not cls.inferred_once


isA = IsA()


@dataclass(eq=False)
class Has(Predicate):
    """
    A predicate that checks if an object type has a certain member object type.
    """
    @classmethod
    def relation(cls):
        return Relation.has

    @classmethod
    def infer(cls, owner_type: Type[TrackedObjectMixin], member_type: Type[TrackedObjectMixin]):
        TrackedObjectMixin.make_class_dependency_graph()

    @classmethod
    def _should_infer(cls) -> bool:
        return not cls.inferred_once


has = Has()


@dataclass(eq=False)
class DependsOn(Predicate):
    """
    A predicate that checks if an object type depends on another object type.
    """

    @classmethod
    def relation(cls):
        return Relation.dependsOn

    @classmethod
    def infer(cls, dependent: Type[TrackedObjectMixin],
                 dependency: Type[TrackedObjectMixin], recursive: bool = False) -> bool:
        raise NotImplementedError("Should be overridden in rdr meta")

    @classmethod
    def _should_infer(cls) -> bool:
        return not cls.inferred_once


dependsOn = DependsOn()
